from __future__ import annotations

import random
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Final
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import InvalidPage, Paginator
from django.http import Http404

from .base import Block, Widget
from .buttons import FormButton
from .headers import BasicHeader, CardHeader
from .text import HTMLWidget

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet


class TabConfig:
    """
    Used to configure the tabs in a :py:class:`PageTabbedWidget`.

    This class holds configuration for an individual tab within a tabbed interface,
    including its name, active state, and optional link.

    Args:
        name: The display name of the tab
        widget: Optional block to display when this tab is active
        active: Whether this tab should be initially active
        link: Optional URL for linking to another page

    Attributes:
        name: The display name of the tab
        active: Whether this tab is active (selected)
        link: URL for the tab if it links to another page

    """

    def __init__(
        self,
        name: str,
        widget: Block | None = None,  # noqa: ARG002
        active: bool = False,
        link: str | None = None,
    ):
        self.name = name
        # TODO: why are we not using widget here?
        # self.widget = widget
        self.active = active
        self.link = link


class PageTabbedWidget(Block):
    """
    Implements a `Tabler Tabbed Card <https://preview.tabler.io/docs/cards.html>`_.

    This widget creates a tabbed interface where the active tab displays a widget,
    and other tabs are links to other pages. When those links are followed, the
    corresponding tab becomes active on the new page. This way, only the active
    widget needs to be rendered.

    Example:
        .. code-block:: python

            from wildewidgets import PageTabbedWidget, TabConfig, Block

            tab = PageTabbedWidget()
            tab.add_link_tab('My First Tab', 'my_first_page_url')
            tab.add_link_tab('My Second Tab', 'my_second_page_url')
            tab.add_active_tab('My Third Tab', widget)

    Args:
        *blocks: Child blocks to include in the tabbed interface

    Keyword Args:
        slug_suffix: Optional suffix for the slug in the URL to differentiate
            between multiple instances of this widget on the same page
        overflow: CSS overflow attribute for the widget (default is "auto")
        **kwargs: Additional keyword arguments passed to the parent class

    """

    #: The Django template to use for rendering this widget.
    template_name: str = "wildewidgets/page_tab_block.html"
    #: The name of the block in the template for CSS styling.
    block: str = "card"
    #: The suffix to use for the slug in the URL.  This is used to
    #: differentiate between multiple instances of this widget on the same page.
    #: This is set to a random number between 0 and 10000 if not provided.
    slug_suffix: str | None = None
    #: The CSS overflow attribute for the widget.
    overflow: str = "auto"

    def __init__(
        self,
        *blocks,
        slug_suffix: str | None = None,
        overflow: str | None = None,
        **kwargs,
    ):
        self.slug_suffix = slug_suffix or self.slug_suffix
        self.overflow = overflow or self.overflow
        super().__init__(*blocks, **kwargs)
        if "style" in self._attributes:
            self._attributes["style"] += f" overflow: {self.overflow};"
        else:
            self._attributes["style"] = f"overflow: {self.overflow};"
        self.tabs: list[TabConfig] = []
        self.widget: Block | None = None

    def add_link_tab(self, name: str, url: str) -> None:
        """
        Add a tab that links to another page.

        Creates a tab that, when clicked, navigates to the specified URL.
        This is used for tabs that are not currently active.

        Args:
            name: The display name of the tab
            url: The URL to navigate to when the tab is clicked

        """
        tab = TabConfig(name=name, link=url)
        self.tabs.append(tab)

    def add_active_tab(self, name: str, widget: Block) -> None:
        """
        Add a tab that displays a widget.

        Creates a tab that is initially active and displays the specified widget.

        Args:
            name: The display name of the tab
            widget: The widget to display when this tab is active

        """
        tab = TabConfig(name=name, active=True)
        self.tabs.append(tab)
        self.widget = widget

    def get_context_data(self, *args, **kwargs):
        """
        Prepare the context data for template rendering.

        Adds the tabs, widget, and unique identifier to the template context.

        Args:
            *args: Positional arguments passed to parent method
            **kwargs: Keyword arguments passed to parent method

        Returns:
            dict: The updated context dictionary with tab-specific data

        """
        kwargs["tabs"] = self.tabs
        if not self.slug_suffix:
            self.slug_suffix = random.randrange(0, 10000)  # noqa: S311
        kwargs["identifier"] = self.slug_suffix
        kwargs["widget"] = self.widget
        return super().get_context_data(*args, **kwargs)


class TabbedWidget(Block):
    """
    Implements a `Tabler Tabbed Card <https://preview.tabler.io/docs/cards.html>`_.

    This widget creates a tabbed interface where all tabs are rendered client-side
    (unlike PageTabbedWidget which only renders the active tab). The tabs are switched
    using JavaScript, without page navigation.

    Example:
        .. code-block:: python

            from wildewidgets import TabbedWidget, Block

            widget1 = Block("This is the first widget")
            widget2 = Block("This is the second widget")
            tab = TabbedWidget()
            tab.add_tab('My First Widget', widget1)
            tab.add_tab('My Second Widget', widget2)

    Args:
        *blocks: Child blocks to include in the tabbed interface

    Keyword Args:
        slug_suffix: Optional suffix for the slug in the URL to differentiate
            between multiple instances of this widget on the same page
        overflow: CSS overflow attribute for the widget (default is "auto")
        **kwargs: Additional keyword arguments passed to the parent class

    """

    #: The Django template to use for rendering this widget.
    template_name: str = "wildewidgets/tab_block.html"
    #: The name of the block in the template for CSS styling.
    block: str = "card"

    #: The suffix to use for the slug in the URL.  This is used to
    #: differentiate between multiple instances of this widget on the same page.
    #: If not provided, a random number between 0 and 10000 will be used.
    #: This is used to ensure that the tabs are unique on the page.
    slug_suffix: str | None = None
    #: The CSS overflow attribute for the widget.
    overflow: str = "auto"

    def __init__(
        self,
        *blocks,
        slug_suffix: str | None = None,
        overflow: str | None = None,
        **kwargs,
    ):
        self.slug_suffix = slug_suffix or self.slug_suffix
        self.overflow = overflow or self.overflow
        super().__init__(*blocks, **kwargs)
        if "style" in self._attributes:
            self._attributes["style"] += f" overflow: {self.overflow};"
        else:
            self._attributes["style"] = f"overflow: {self.overflow};"
        #: A list of tuples containing the name of the tab and the widget to
        #: display in that tab.  Each tuple is of the form (name, widget), where
        #: `name` is a string and `widget` is a Block.
        self.tabs: list[tuple[str, Block]] = []

    def add_tab(self, name: str, widget: Block) -> None:
        """
        Add a tab to the tabbed interface.

        Creates a tab with the specified name that displays the specified widget
        when the tab is selected.

        Args:
            name: The display name of the tab
            widget: The widget to display when this tab is selected

        """
        self.tabs.append((name, widget))

    def get_context_data(self, *args, **kwargs):
        """
        Prepare the context data for template rendering.

        Adds the tabs and unique identifier to the template context.

        Args:
            *args: Positional arguments passed to parent method
            **kwargs: Keyword arguments passed to parent method

        Returns:
            dict: The updated context dictionary with tab-specific data

        """
        kwargs["tabs"] = self.tabs
        if not self.slug_suffix:
            self.slug_suffix = random.randrange(0, 10000)  # noqa: S311
        kwargs["identifier"] = self.slug_suffix
        # TODO: why is overflow not in kwargs?
        # kwargs['overflow'] = self.overflow
        return super().get_context_data(*args, **kwargs)


class CardWidget(Block):
    """
    Renders a `Bootstrap 5 Card <https://getbootstrap.com/docs/5.2/components/card/>`_.

    This widget creates a Bootstrap card component with optional header, title,
    subtitle, and body content.

    Attributes:
        template_name: The Django template to use for rendering this widget
        block: The name of the block in the template for CSS styling
        header: Optional header widget to display at the top of the card
        header_text: Optional text to use for a simple header (if header is not
            provided)
        header_css: Optional CSS classes to apply to the header
        card_title: Optional title for the card
        card_subtitle: Optional subtitle for the card
        widget: The main widget to display in the card body (required)
        widget_css: Optional CSS classes to apply to the widget
        overflow: CSS overflow attribute for the card

    Example:
        .. code-block:: python

            from wildewidgets import CardWidget, Block

            # Simple card with just a widget
            card = CardWidget(
                widget=Block("Card content"),
                card_title="My Card",
                card_subtitle="Card subtitle"
            )

            # Card with a header and custom widget CSS
            card = CardWidget(
                header_text="Card Header",
                widget=Block("Card content"),
                widget_css="p-4"
            )

    Args:
        *blocks: Child blocks to include in the card body

    Keyword Args:
        header: Optional header widget to display at the top of the card
        header_text: Optional text to use for a simple header (if header is not
            provided)
        header_css: Optional CSS classes to apply to the header
        card_title: Optional title for the card
        card_subtitle: Optional subtitle for the card
        widget: The main widget to display in the card body (required)
        widget_css: Optional CSS classes to apply to the widget
        overflow: CSS overflow attribute for the card

    """

    template_name: str = "wildewidgets/card_block.html"

    block: str = "card"

    header: BasicHeader | None = None
    header_text: str | None = None
    header_css: str | None = None
    card_title: str | None = None
    card_subtitle: str | None = None
    widget: Widget | None = None
    widget_css: str | None = None
    overflow: str = "auto"

    def __init__(
        self,
        *blocks,
        header: BasicHeader | None = None,
        header_text: str | None = None,
        header_css: str | None = None,
        card_title: str | None = None,
        card_subtitle: str | None = None,
        widget: Widget | None = None,
        widget_css: str | None = None,
        overflow: str | None = None,
        **kwargs,
    ):
        self.header = header or deepcopy(self.header)
        self.header_text = header_text or self.header_text
        self.header_css = header_css or self.header_css
        if self.header_text and not self.header:
            self.header = CardHeader(header_text=self.header_text)
        self.card_title = card_title or self.card_title
        self.card_subtitle = card_subtitle or self.card_subtitle
        self.widget = widget or deepcopy(self.widget)
        self.overflow = overflow or self.overflow
        self.widget_css = widget_css or self.widget_css
        super().__init__(*blocks, **kwargs)
        if "style" in self._attributes:
            self._attributes["style"] += f" overflow: {self.overflow};"
        else:
            self._attributes["style"] = f"overflow: {self.overflow};"

    def get_context_data(self, *args, **kwargs):
        """
        Prepare the context data for template rendering.

        Adds the card components (header, title, subtitle, widget) to the context.

        Args:
            *args: Positional arguments passed to parent method
            **kwargs: Keyword arguments passed to parent method

        Returns:
            dict: The updated context dictionary with card-specific data

        Raises:
            ImproperlyConfigured: If no widget is defined for the card

        """
        kwargs = super().get_context_data(*args, **kwargs)
        kwargs["header"] = self.header
        kwargs["header_text"] = self.header_text
        kwargs["header_css"] = self.header_css
        kwargs["title"] = self.card_title
        kwargs["subtitle"] = self.card_subtitle
        kwargs["widget"] = self.widget
        kwargs["widget_css"] = self.widget_css
        kwargs["css_class"] = self.css_class

        if not self.widget:
            msg = "You must define a widget."
            raise ImproperlyConfigured(msg)
        return kwargs

    def set_widget(self, widget, css_class=None):
        """
        Set or replace the main widget displayed in the card body.

        Args:
            widget: The widget to display in the card body
            css_class: Optional CSS classes to apply to the widget

        """
        self.widget = widget
        self.widget_css = css_class

    def set_header(self, header):
        """
        Set or replace the header widget displayed at the top of the card.

        Args:
            header: The header widget to display

        """
        self.header = header


class MultipleModelWidget(Block):
    """
    Base class for widgets that display multiple model instances.

    This abstract class provides common functionality for widgets that display
    a list of model instances, such as PagedModelWidget and ListModelWidget.

    Note:
        Either model or queryset must be provided, but not both.

    Example:
        .. code-block:: python

            from wildewidgets import MultipleModelWidget
            from myapp.models import MyModel
            from myapp.widgets import MyModelWidget

            class MyListWidget(MultipleModelWidget):
                model = MyModel
                model_widget = MyModelWidget
                ordering = "-created_at"
                item_label = "item"

    Args:
        *blocks: Child blocks to include in the widget

    Keyword Args:
        model: The Django model class to query for instances
        queryset: A pre-defined queryset to use for fetching model instances
        ordering: The ordering to apply to the queryset (default is None)
        model_widget: The widget class to use for rendering each model instance
        model_kwargs: Optional keyword arguments to pass to the model widget
        item_label: The label to use for each instance (default is "item")

    Raises:
        ImproperlyConfigured: If both model and queryset are defined, or if neither
            is defined, or if model_widget is not defined.

    """

    #: If this is defined, do ``self.model.objects.all()`` to get our list of instances.
    #: Define either this or :py:attr:`queryset`, but not both.
    model: type[Model] | None = None
    #: Use this queryset to get our list of instances.  Define either this or
    #: :py:attr:`model`, but not both.
    queryset: QuerySet | None = None
    #: The ordering to use for the queryset.  This will be applied to both
    #: :py:attr:`model` and :py:attr:`queryset`
    ordering: str | tuple[str, ...] | None = None

    #: Use this widget class to render each model instance.
    model_widget: type[Widget] | None = None
    #: When instantiating :py:attr:`model_widget`, pass this dict into the
    #: widget constructor as the keyword arguments
    model_kwargs: dict[str, Any] = {}  # noqa: RUF012

    #: The label to use for each instance.  This is used in the confirmation
    #: dialog when deleting instances.
    item_label: str = "item"

    def __init__(
        self,
        *blocks,
        model: type[Model] | None = None,
        queryset: QuerySet | None = None,
        ordering: str | None = None,
        model_widget: type[Widget] | None = None,
        model_kwargs: dict[str, Any] | None = None,
        item_label: str | None = None,
        **kwargs,
    ) -> None:
        self.model = model or self.model
        self.queryset = queryset if queryset is not None else self.queryset
        if self.model and self.queryset:
            msg = "You must define either model or queryset, but not both."
            raise ImproperlyConfigured(msg)
        self.model_widget = model_widget or deepcopy(self.model_widget)
        self.model_kwargs = model_kwargs or deepcopy(self.model_kwargs)
        self.ordering = ordering or self.ordering
        self.item_label = item_label or self.item_label
        super().__init__(*blocks, **kwargs)

    def get_item_label(self, instance: Model) -> str:  # noqa: ARG002
        """
        Get the label to use for a specific model instance.

        This label is used in confirmation dialogs and other UI elements.

        Args:
            instance: The model instance to get the label for

        Returns:
            str: The label for the instance (defaults to the class's item_label)

        """
        # TODO: why is instance an argument here?  It is not used here.
        return self.item_label

    def get_model_widget(self, object: Model, **kwargs) -> Widget:  # noqa: A002
        """
        Create a widget for a specific model instance.

        This method creates and returns a widget for displaying a single model instance.

        Args:
            object: The model instance to create a widget for
            **kwargs: Additional keyword arguments to pass to the widget constructor

        Returns:
            Widget: The widget for displaying the model instance

        Raises:
            ImproperlyConfigured: If model_widget is not defined and this method
                is not overridden

        """
        if self.model_widget:
            return self.model_widget(object=object, **kwargs)
        msg = (
            f"{self.__class__.__name__} is missing a model widget. Define "
            f"{self.__class__.__name__}.model_widget or override "
            f"{self.__class__.__name__}.get_model_widget()."
        )
        raise ImproperlyConfigured(msg)

    def get_model_widgets(self, instances: list[Model]) -> list[Widget]:
        """
        Create widgets for a list of model instances.

        Args:
            instances: List of model instances to create widgets for

        Returns:
            list[Widget]: List of widgets for displaying the model instances

        """
        return [
            self.get_model_widget(object=instance, **self.model_kwargs)
            for instance in instances
        ]

    def get_queryset(self) -> list[Model] | QuerySet:
        """
        Get the queryset of model instances to display.

        This method fetches the model instances to display, either from the
        provided queryset or by querying the model.

        Returns:
            list[Model] | QuerySet: The model instances to display

        Raises:
            ImproperlyConfigured: If neither model nor queryset is defined

        """
        if self.queryset is not None:
            queryset = self.queryset
            queryset = queryset.all()
        elif self.model is not None:
            # TODO: why are we using _default_manager here instead of .objects ?
            # The class may have replaced .objects with a more targeted manager and
            # we won't get that if we always hit _default_manager
            queryset = self.model._default_manager.all()
        else:
            msg = (
                f"{self.__class__.__name__} is missing a QuerySet. Define "
                f"{self.__class__.__name__}.model, {self.__class__.__name__}.queryset, "
                f"or override {self.__class__.__name__}.get_queryset()."
            )
            raise ImproperlyConfigured(msg)
        ordering = self.ordering
        if ordering:
            if isinstance(ordering, str):
                ordering = (ordering,)
            queryset = queryset.order_by(*ordering)
        return queryset


class PagedModelWidget(MultipleModelWidget):
    """
    A widget that displays a paginated list of model instances.

    This widget creates a paginated list of widgets, with each widget displaying
    a single model instance. It includes pagination controls for navigating
    between pages.

    Example:
        .. code-block:: python

            from wildewidgets import PagedModelWidget, MyObjectWidget

            widget = PagedModelWidget(
                queryset=mymodel.myobject_set.all(),
                paginate_by=10,
                page_kwarg='page',
                model_widget=MyObjectWidget,
                model_kwargs={'foo': 'bar'},
                item_label="object",
                extra_url={
                    'pk': mymodel.id,
                    '#': 'objects',
                },
            )

    Args:
        *blocks: Child blocks to include in the widget

    Keyword Args:
        model: The Django model class to query for instances (optional)
        queryset: A pre-defined queryset to use for fetching model instances
        ordering: The ordering to apply to the queryset (default is None)
        model_widget: The widget class to use for rendering each model instance
        model_kwargs: Optional keyword arguments to pass to the model widget
        item_label: The label to use for each instance (default is "item")
        extra_url: Additional URL parameters to include in pagination links
        page_kwarg: The GET parameter name for the page number (default is "page")
        paginate_by: Number of items per page (default is 25)
        max_page_controls: Maximum number of page controls to display (default is 5)s

    """

    #: The Django template to use for rendering this widget.
    template_name: str = "wildewidgets/paged_model_widget.html"
    #: The name of the block in the template for CSS styling.
    block: str = "wildewidgets-paged-model-widget"

    #: The get argument for the page number. Defaults to `page`.
    page_kwarg: str = "page"
    #: Number of widgets per page. Defaults to all widgets.
    paginate_by: int = 25
    #: Extra parameters passed in the page links.
    max_page_controls: int = 5

    def __init__(
        self,
        *blocks,
        page_kwarg: str | None = None,
        paginate_by: int | None = None,
        extra_url: dict[str, Any] | None = None,
        **kwargs,
    ):
        self.page_kwarg = page_kwarg or self.page_kwarg
        self.paginate_by = paginate_by or self.paginate_by
        self.extra_url = extra_url or {}
        super().__init__(*blocks, **kwargs)

    def get_context_data(self, *args, **kwargs):
        """
        Prepare the context data for template rendering.

        This method:
        1. Handles pagination of the model instances
        2. Creates widgets for the current page of instances
        3. Prepares pagination controls and context
        4. Adds any extra URL parameters for pagination links

        Args:
            *args: Positional arguments passed to parent method
            **kwargs: Keyword arguments passed to parent method

        Returns:
            dict: The updated context dictionary with pagination-specific data

        Raises:
            Http404: If an invalid page number is requested

        """
        kwargs = super().get_context_data(*args, **kwargs)
        self.request = kwargs.get("request")
        if self.paginate_by:
            paginator = Paginator(self.get_queryset(), self.paginate_by)
            page_number = self.request.GET.get(self.page_kwarg)
            try:
                page_number = int(page_number)
            except (TypeError, ValueError):
                page_number = 1
            try:
                page = paginator.page(page_number)
            except InvalidPage as e:
                msg = f"Invalid page ({page_number}): {e!s}"
                raise Http404(msg) from e
            kwargs["widget_list"] = self.get_model_widgets(page.object_list)
            kwargs["page_obj"] = page
            kwargs["is_paginated"] = page.has_other_pages()
            kwargs["paginator"] = paginator
            kwargs["page_kwarg"] = self.page_kwarg
            pages = kwargs["page_obj"].paginator.num_pages
            pages = min(pages, self.max_page_controls)
            page_number = page.number
            max_controls_half = int(self.max_page_controls / 2)
            range_start = max(page_number - max_controls_half, 1)
            kwargs["page_range"] = range(range_start, range_start + pages)
        else:
            kwargs["widget_list"] = self.get_model_widgets(self.get_queryset().all())
        kwargs["item_label"] = self.item_label
        if self.extra_url:
            anchor = self.extra_url.pop("#", None)
            extra_url = f"&{urlencode(self.extra_url)}"
            if anchor:
                extra_url = f"{extra_url}#{anchor}"
            kwargs["extra_url"] = extra_url
        else:
            kwargs["extra_url"] = ""
        return kwargs


class CollapseWidget(Block):
    """
    A `Bootstrap Collapse widget <https://getbootstrap.com/docs/5.2/components/collapse/>`_.

    This widget creates a collapsible content area that can be toggled open or
    closed by a CollapseButton or other trigger element. It's useful for hiding
    content that isn't immediately relevant but might be needed later.

    Note:
        A `CollapseWidget` needs a trigger element (like a
        :py:class:`wildewidgets.CollapseButton`) with the `data-toggle="collapse"`
        attribute and a `data-target` pointing to this widget's ID.

    Example:
        .. code-block:: python

            from wildewidgets import CollapseWidget, CardHeader, CrispyFormWidget

            collapse_id = 'my-collapse-id'
            collapse = CollapseWidget(
                CrispyFormWidget(form=form),
                css_id=collapse_id,
                css_class="pt-3",
            )

            header = CardHeader(header_text="Services")
            header.add_collapse_button(
                text="Manage",
                color="primary",
                target=f"#{collapse_id}",
            )

    """

    #: The Django template to use for rendering this widget.
    block: str = "collapse"


class HorizontalLayoutBlock(Block):
    """
    A container that aligns child widgets horizontally using flexbox.

    This widget creates a horizontal layout for its child widgets using Bootstrap's
    flexbox utilities. It provides options for controlling vertical and horizontal
    alignment, as well as responsive behavior.

    Example:
        .. code-block:: python

            from wildewidgets import HorizontalLayoutBlock, LabelBlock, Block

            # Create a layout with right-aligned items and centered vertical alignment
            layout = HorizontalLayoutBlock(
                LabelBlock("Label"),
                Block("Content 1"),
                Block("Content 2"),
                justify="end",
                align="center",
                flex_size="md",  # Stack vertically on screens smaller than medium
                css_class="mt-3"
            )

    Args:
        *blocks: Child blocks to include in the horizontal layout

    Keyword Args:
        align: Vertical alignment of items
        justify: Horizontal alignment of items
        flex_size: Bootstrap viewport size below which items stack vertically
        **kwargs: Additional keyword arguments passed to the parent class

    """

    #: The valid column content ``justify-content-`` values
    VALID_JUSTIFICATIONS: Final[list[str]] = [
        "start",
        "center",
        "end",
        "between",
        "around",
        "evenly",
    ]
    #: The valid column content ``justify-content-`` values
    VALID_ALIGNMENTS: Final[list[str]] = [
        "start",
        "center",
        "end",
        "baseline",
        "stretch",
    ]

    #: How to align items veritcally within this widget.  Valid choices: ``start``,
    #: ``center``, ``end``, ``baselin``, ``stretch``.  See `Bootstrap: Flex,
    #: justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#align-items>`_.
    #: If not supplied here and :py:attr:`align` is ``None``, do whatever
    #: vertical aligment Bootstrap does by default.
    align: str = "center"
    #: How to align items horizontally within this widget.  Valid choices:
    #: ``start``, : ``center``, ``end``, ``between``, ``around``, ``evenly``.  See
    #: `Bootstrap: Flex, justify content
    #: <https://getbootstrap.com/docs/5.2/utilities/flex/#justify-content>`_.
    #: If not supplied here and :py:attr:`justify` is ``None``, do whatever
    #: horizontal aligment Bootstrap does by default.
    justify: str = "between"
    #: the Boostrap viewport size below which this will be a vertical list instead
    #: of a horizontal one.
    flex_size: str | None = None

    def __init__(
        self,
        *blocks,
        align: str | None = None,
        justify: str | None = None,
        flex_size: str | None = None,
        **kwargs,
    ):
        self.align = align or self.align
        self.justify = justify or self.justify
        self.flex_size = flex_size or self.flex_size
        if self.align not in self.VALID_ALIGNMENTS:
            msg = (
                f'"{self.align}" is not a valid vertical alignment value.  Valid '
                f"values are {', '.join(self.VALID_ALIGNMENTS)}"
            )
            raise ValueError(msg)
        if self.justify not in self.VALID_JUSTIFICATIONS:
            msg = (
                f'"{self.justify}" is not a valid horizontal alignment value.  Valid '
                f"values are {', '.join(self.VALID_JUSTIFICATIONS)}"
            )
            raise ValueError(msg)
        super().__init__(*blocks, **kwargs)
        flex = f"d-{self.flex_size}-flex" if self.flex_size else "d-flex"
        self.add_class(flex)
        self.add_class(f"align-items-{self.align}")
        self.add_class(f"justify-content-{self.justify}")


class ListModelWidget(MultipleModelWidget):
    """
    A widget that displays a list of model instances.

    This widget creates an unordered list of items, with each item displaying a
    single model instance. It can optionally include buttons for removing items.

    Example:
        .. code-block:: python

            from wildewidgets import ListModelWidget
            from django.urls import reverse

            # Create a list with remove buttons
            widget = ListModelWidget(
                queryset=parent.children.all(),
                item_label='child',
                remove_url=reverse('remove_child') + "?id={}",
            )

            # Create a read-only list with custom model widget
            widget = ListModelWidget(
                queryset=Author.objects.all(),
                model_widget=AuthorWidget,
                item_label='author'
            )

    Args:
        *args: Positional arguments passed to parent class

    Keyword Args:
        remove_url: Optional URL format string for removing items (with {}
            placeholder for ID)
        **kwargs: Keyword arguments passed to parent class

    Raises:
        ValueError: If model_widget is not defined and no model instance is provided
        ImproperlyConfigured: If neither model nor queryset is defined

    """

    #: The name of the block in the template for CSS styling.
    block: str = "list-group"
    #: Another name for this widget, used in addition to the block name.
    name: str = "wildewidgets-list-model-widget"
    #: The HTML tag to use for the container
    tag: str = "ul"
    #: If True, show a message when there are no items in the list.
    show_no_items: bool = True

    #: The url to "POST" to in order to delete or remove the object.
    remove_url: str | None = None

    def __init__(self, *args, remove_url: str | None = None, **kwargs: Any) -> None:
        self.remove_url = remove_url or self.remove_url
        super().__init__(*args, **kwargs)
        result = self.get_queryset()
        if not isinstance(result, list):
            result = list(result.all())
        widgets = self.get_model_widgets(result)
        if not widgets and self.show_no_items:
            self.add_block(
                Block(
                    f"No {self.item_label}s",
                    tag="li",
                    name="list-group-item",
                    css_class="fw-light fst-italic border",
                )
            )
        for widget in widgets:
            self.add_block(widget)

    def get_remove_url(self, instance: Model) -> str:
        """
        Get the URL for removing a specific model instance.

        Args:
            instance: The model instance to get the remove URL for

        Returns:
            str: The URL for removing the instance, or an empty string if not configured

        """
        if self.remove_url:
            return self.remove_url.format(instance.id)
        return ""

    def get_confirm_text(self, instance: Model) -> str:
        """
        Get the confirmation text for removing a specific model instance.

        This text is used in the confirmation dialog when removing an item.

        Args:
            instance: The model instance to get the confirmation text for

        Returns:
            str: The confirmation text for removing the instance

        """
        item_label = self.get_item_label(instance)
        return f"Are you sure you want to remove this {item_label}?"

    def get_object_text(self, instance: Model) -> str:
        """
        Get the display text for a specific model instance.

        Args:
            instance: The model instance to get the display text for

        Returns:
            str: The display text for the instance (defaults to str(instance))

        """
        return str(instance)

    def get_model_subblock(self, instance: Model):
        """
        Create a block for displaying a model instance.

        If the model instance has a get_absolute_url method, the text will be
        wrapped in a link to that URL.

        Args:
            instance: The model instance to create a block for

        Returns:
            Block: A block for displaying the model instance

        """
        if hasattr(instance, "get_absolute_url"):
            url = instance.get_absolute_url()
            # TODO: use a Link here
            return Block(
                HTMLWidget(
                    html=f'<a href="{url}"><label>{self.get_object_text(instance)}'
                    "</label></a>"
                )
            )
        return Block(self.get_object_text(instance), tag="label")

    def get_model_widget(self, object: Model, **kwargs) -> Widget:  # type: ignore[override]  # noqa: A002
        """
        Create a widget for a specific model instance.

        If model_widget is defined, it uses that. Otherwise, it creates a default
        widget with the instance text and an optional remove button.

        Args:
            object: The model instance to create a widget for

        Keyword Args:
            **kwargs: Additional keyword arguments for the widget

        Returns:
            Widget: The widget for displaying the model instance

        Raises:
            ValueError: If obj is None and model_widget is not defined

        """
        if self.model_widget:
            return super().get_model_widget(object=object, **kwargs)
        if object is None:
            if not self.model_widget:
                msg = (
                    f"{self.__class__.__name__} is missing a model widget. Define "
                    f"{self.__class__.__name__}.model_widget or override "
                    f"{self.__class__.__name__}.get_model_widget()."
                )
                raise ValueError(msg)
            return self.model_widget(**kwargs)
        widget = HorizontalLayoutBlock(
            tag="li", name="list-group-item listmodelwidget__item"
        )
        widget.add_block(self.get_model_subblock(object))
        remove_url = self.get_remove_url(object)
        if remove_url:
            widget.add_block(
                FormButton(
                    close=True,
                    action=remove_url,
                    confirm_text=self.get_confirm_text(object),
                ),
            )
        return widget


class ListModelCardWidget(CardWidget):
    """
    A card widget containing a filterable list of model instances.

    This widget creates a card with a header containing a filter input field
    and a body containing a list of model instances. The filter input allows
    users to filter the list by typing.

    Example:
        .. code-block:: python

            from wildewidgets import ListModelCardWidget
            from myapp.models import Author
            from myapp.widgets import AuthorListWidget

            # Basic usage with default list widget
            widget = ListModelCardWidget(
                queryset=Author.objects.all()[:10]
            )

            # Custom list widget and placeholder
            widget = ListModelCardWidget(
                queryset=Author.objects.all(),
                list_model_widget_class=AuthorListWidget,
                placeholder="Search authors...",
                item_label="author"
            )

    Args:
        *args: Positional arguments passed to parent class

    Keyword Args:
        list_model_widget_class: Optional custom widget class for the list model
        list_model_header_class: Optional custom header widget class
        placeholder: Placeholder text for the filter input field
        **kwargs: Keyword arguments passed to parent class

    """

    #: The script to use for filtering the list of objects.
    SCRIPT: str = """
var filter_input = document.getElementById("{filter_id}");
filter_input.onkeyup = function(e) {{
    var filter = e.target.value.toLowerCase();
    document.querySelectorAll("{query} label").forEach(label => {{
        var test_string = label.innerText.toLowerCase();
        if (test_string.includes(filter)) {{
            label.closest('.listmodelwidget__item').classList.remove('d-none');
        }}
        else {{
            label.closest('.listmodelwidget__item').classList.add('d-none');
        }}
    }});
    let children = document.querySelectorAll("{query} li");
    for (let i=0; i < children.length; i++) {{
        let child = children[i];
        child.classList.remove('border-top');
    }};
    for (let i=0; i < children.length; i++) {{
        let child = children[i];
        if (child.classList.contains('d-none')) {{
        }}
        else {{
            child.classList.add('border-top');
            break;
        }}
    }};
}};

"""

    #: The Widget subclass to use for the list model widget.
    list_model_widget_class: type[Widget] = ListModelWidget
    #: The Widget subclass to use for the header.
    list_model_header_class: type[Widget] | None = None

    def __init__(
        self,
        *args,
        list_model_widget_class: type[Widget] | None = None,
        list_model_header_class: type[Widget] | None = None,
        placeholder: str | None = None,
        **kwargs,
    ):
        self.id_base = f"list_modal_card_{random.randrange(0, 1000)}"  # noqa: S311
        self.list_model_widget_id = f"{self.id_base}_list_model_widget"
        self.filter_id = f"{self.id_base}_filter"
        self.list_model_widget_class = (
            list_model_widget_class or self.list_model_widget_class
        )
        self.list_model_header_class = (
            list_model_header_class or self.list_model_header_class
        )
        # Pop the kwargs that are used to build the widget and header.
        # These will be passed to the widget and header constructors.
        widget_kwargs = {
            "remove_url": kwargs.pop("remove_url", None),
            "model": kwargs.pop("model", None),
            "model_widget": kwargs.pop("model_widget", None),
            "ordering": kwargs.pop("ordering", None),
            "queryset": kwargs.pop("queryset", None),
            "model_kwargs": kwargs.pop("model_kwargs", {}),
            "item_label": kwargs.pop("item_label", "item"),
            "css_id": self.list_model_widget_id,
        }
        self.widget: Widget = self.get_list_model_widget(**widget_kwargs)
        self.placeholder = placeholder or f"Filter {self.widget.item_label}s"
        kwargs["widget"] = self.widget
        header_kwargs = {
            "title": kwargs.pop("title", ""),
        }
        kwargs["header"] = self.get_list_model_header(**header_kwargs)
        kwargs["header_css"] = "bg-light"
        filter_label_query = f"#{self.list_model_widget_id}"
        kwargs["script"] = self.SCRIPT.format(
            query=filter_label_query, filter_id=self.filter_id
        )
        super().__init__(*args, **kwargs)

    def get_list_model_widget(self, *args, **kwargs):
        """
        Create the list model widget.

        Args:
            *args: Positional arguments for the list model widget
            **kwargs: Keyword arguments for the list model widget

        Returns:
            Widget: The list model widget instance

        """
        return self.list_model_widget_class(*args, **kwargs)

    def get_list_model_header(self, *args, **kwargs):
        """
        Create the header widget with filter input.

        If list_model_header_class is defined, it uses that. Otherwise, it creates
        a default header with a filter input field.

        Args:
            *args: Positional arguments for the header
            **kwargs: Keyword arguments for the header

        Returns:
            Widget: The header widget instance

        """
        from .forms import InputBlock, LabelBlock

        if self.list_model_header_class:
            return self.list_model_header_class(*args, **kwargs)  # pylint: disable=not-callable
        return Block(
            Block(
                LabelBlock(
                    f"Filter {self.widget.item_label}s",
                    css_class="d-none",
                    for_input=self.filter_id,
                ),
                InputBlock(
                    attributes={
                        "type": "text",
                        "placeholder": self.placeholder,
                    },
                    css_id=f"{self.id_base}_filter",
                    css_class="form-control",
                ),
                css_class="w-25",
            ),
            Block(""),
            css_class="d-flex flex-row-reverse w-100",
        )



from __future__ import annotations

import re
from copy import deepcopy
from functools import partial
from typing import Any, Final, cast

from .base import Block


class Column(Block):
    """
    Implements a ``col`` from the `Bootstrap Grid system
    <https://getbootstrap.com/docs/5.2/layout/grid/>`_.

    This is primarily meant to be used with :py:class:`Row`, but you can
    actually use :py:class:`Column` objects outside of a :py:class:`Row`.  This
    allows you to specify a particular width for the block within a page.  See
    `Boostrap: Columns, Standalone Column Classes
    <https://getbootstrap.com/docs/5.2/layout/columns/#standalone-column-classes>`_.

    Example:
        .. code-block:: python

            from wildewidgets import Column, Block

            col = Column(
                Block("This goes at the top of the column", css_class="mb-2"),
                Block("This goes at the bottom of the column"),
                base_width=6,
                viewport_widths={"md": "4", "lg": "3"},
                alignment="center",
                self_alignment="center"
            )

    Args:
        *blocks: a list of :py:class:`Block` objects or string to add to this column

    Keyword Args:
        base_width: the base width of this block for all viewports.  This will
            be converted to a ``col-{base_width}`` CSS class.  If this is not
            supplied and :py:attr:`base_width` is ``None``, add ``col`` as our
            CSS class.
        viewport_widths: a dict where the keys are Bootstrap viewport sizes (e.g.
            ``sm``, ``xl``) and the values are column widths, again between 1
            and 12 inclusive, or "auto".  These will be converted to CSS classes
            that look like ``col-{viewport}-{width}``
        alignment: how to align items within this column.  Valid choices: ``start``,
            ``center``, ``end``, ``between``, ``around``, ``evenly``.  See
            `Bootstrap: Flex, justify content
            <https://getbootstrap.com/docs/5.2/utilities/flex/#justify-content>`_.
            If not supplied here and :py:attr:`alignment` is ``None``, do
            whatever aligment Bootstrap does by default.
        self_alignment: how to align this column vertically within its
            containing row.  Valid choices: ``start``, ``end``, ``center``  See
            `Bootstrap Columns
            <https://getbootstrap.com/docs/5.2/layout/columns/#alignment>`_

    Raises:
        ValueError: there was a problem validating one of our settings

    """

    #: The minimum and maximum column widths, as per Bootstrap
    COL_MIN_WIDTH: Final[int] = 1
    COL_MAX_WIDTH: Final[int] = 12

    #: The valid column content alignments
    ALIGNMENTS: Final[list[str]] = [
        "start",
        "center",
        "end",
        "between",
        "around",
        "evenly",
    ]
    #: The valid self vertical alignments with our row
    SELF_ALIGNMENTS: Final[list[str]] = ["start", "center", "end"]

    #: A column width between 1 and 12 inclusive.  This is the base width
    #: for all viewports
    base_width: int | None = None
    #: A dict where the keys are Bootstrap viewport sizes (e.g.  #: ``sm``,
    #: ``xl``) and the values are column widths, again between 1 and 12
    #: inclusive, or "auto"
    viewport_widths: dict[str, str] = {}  # noqa: RUF012
    #: How to align items within this column.  Valid choices: ``start``,
    #: ``center``, ``end``, ``between``, ``around``, ``evenly``.  See `Bootstrap: Flex,
    #: justify content <https://getbootstrap.com/docs/5.2/utilities/flex/#justify-content>`_.
    #: If not supplied here and :py:attr:`alignment` is ``None``, do whatever aligment
    #: Bootstrap does by default.
    alignment: str | None = None
    #: How to align this column vertically within its containing row.  Valid
    #: choices: ``start``, ``end``, ``center``.  See `Bootstrap: Columns,
    #: Alignment : <https://getbootstrap.com/docs/5.2/layout/columns/#alignment>`_
    self_alignment: str | None = None

    def __init__(
        self,
        *blocks: Block,
        base_width: int | None = None,
        viewport_widths: dict[str, str] | None = None,
        alignment: str | None = None,
        self_alignment: str | None = None,
        **kwargs: Any,
    ):
        self.base_width = base_width or self.base_width
        self.viewport_widths = viewport_widths or deepcopy(self.viewport_widths)
        self.alignment = alignment or self.alignment
        self.self_alignment = self_alignment or self.self_alignment
        self.check_widths()
        self.check_alignments()
        super().__init__(*blocks, **kwargs)
        if self.base_width:
            self.add_class(f"col-{self.base_width}")
        elif self.viewport_widths:
            self.add_class("col-12")
        else:
            self.add_class("col")
        for viewport, w in cast("dict[str, str]", self.viewport_widths).items():
            self.add_class(f"col-{viewport}-{w}")
        if self.alignment or self.self_alignment:
            self.add_class("d-flex")
        if self.alignment:
            self.add_class(f"justify-content-{self.alignment}")
        if self.self_alignment:
            self.add_class(f"align-self-{self.self_alignment}")

    def check_widths(self) -> None:
        """
        Validate our supplied ``width`` and ``viewport_widths`` settings to
        ensure widths for every viewport are between 0 and 12 inclusive, or (in
        the case of viewport specific widths).

        Raises:
            ValueError: a width was out of range

        """
        # TODO: validate viewport names also
        width: str | int
        if self.base_width:
            try:
                width = int(self.base_width)
            except ValueError as exc:
                msg = f'Invalid width "{self.base_width}".  Width must be an integer.'
                raise ValueError(msg) from exc
            else:
                if width and (width < 1 or width > self.COL_MAX_WIDTH):
                    msg = (
                        f"Invalid width {self.base_width}.  Width must be 0 > "
                        "width <= 12"
                    )
                    raise ValueError(msg)
        if self.viewport_widths:
            for viewport, width in self.viewport_widths.items():
                try:
                    _width = int(width)
                except ValueError as e:  # noqa: PERF203
                    if _width != "auto":
                        msg = (
                            f'Invalid width "{_width}" for viewport "{viewport}".  '
                            'Width must be either "auto" or, an integer 0 > width <= 12'
                        )
                        raise ValueError(msg) from e
                else:
                    if _width and (
                        _width < self.COL_MIN_WIDTH or _width > self.COL_MAX_WIDTH
                    ):
                        msg = (
                            f'Invalid width {width} for viewport "{viewport}".  Width '
                            "must be an integer 0 > width <= 12"
                        )
                        raise ValueError(msg)

    def check_alignments(self) -> None:
        """
        Check that our supplied ``alignment`` and ``self_alignment``  settings
        to ensure they are valid values from :py:attr:`ALIGNMENTS` and
        :py:attr:`SELF_ALIGNMENTS`, respectively.

        Raises:
            ValueError: an alignment was not valid

        """
        if self.alignment and self.alignment not in self.ALIGNMENTS:
            msg = (
                f"Invalid alignment: {self.alignment}.  Alignment must be one of "
                f"{', '.join(self.ALIGNMENTS)}"
            )
            raise ValueError(msg)
        if self.self_alignment and self.self_alignment not in self.SELF_ALIGNMENTS:
            msg = (
                f"Invalid self_alignment: {self.self_alignment}.  "
                f"Alignment must be one of {', '.join(self.SELF_ALIGNMENTS)}"
            )
            raise ValueError(msg)


class Row(Block):
    """
    Implements a ``row`` from the `Bootstrap Grid system
    <https://getbootstrap.com/docs/5.2/layout/grid/>`_.

    As columns are added to this Row, helper methods are also added
    to this :py:class:`row` instance, named for the :py:attr:`Column.name`
    of the column.  See :py:meth:`_add_helper_method` for details on how
    the helper methods will be named.

    Example:
        .. code-block:: python

            from wildewidgets import Row, Column, Block

            sidebar = Column(
                Block("This is in the sidebar"),
                name='sidebar',
                width=3
            )
            main = Column(
                Block("This is in the main content area"),
                name='main'
            )
            row = Row(
                sidebar,
                main,
                horizontal_alignment='center',
                vertical_alignment='center'
            )

    Args:
        *columns: one or more :py:class:`Column` objects

    Keyword Args:
        horizontal_alignment: the horizontal alignment for the columns in this
            row.  See `Bootstrap: Columns, Horizontal Alignment
            <https://getbootstrap.com/docs/5.2/layout/columns/#horizontal-alignment>`_.
            Valid choices: ``start``, ``center``, ``end``, ``around``, ``between``,
            ``evenly``.
        vertical_alignment: the vertical alignment for the columns in this
            row.  See `Bootstrap: Columns, Vertical Alignment
            <https://getbootstrap.com/docs/5.2/layout/columns/#vertical-alignment>`_.
            Valid choices: ``start``, ``center``, ``end``.

    Raises:
        ValueError: there was a problem validating one of our alignments

    """

    HORIZONTAL_ALIGNMENT: Final[list[str]] = [
        "start",
        "center",
        "end",
        "around",
        "between",
        "evenly",
    ]
    VERTICAL_ALIGNMENT: Final[list[str]] = ["start", "center", "end"]

    block: str = "row"

    #: A list of :py:class:`Column` blocks to add to the this row.
    #: This is List[Block] here so people can specify their own Column-like classes
    columns: list[Block] = []  # noqa: RUF012
    #: the horizontal alignment for the columns in this  row.  See
    #: `Bootstrap: Columns, Horizontal Alignment
    #: <https://getbootstrap.com/docs/5.2/layout/columns/#horizontal-alignment>`_
    horizontal_alignment: str | None = None
    #: the vertical alignment for the columns in this  row.  See
    #: `Bootstrap: Columns, Vertical Alignment
    #: <https://getbootstrap.com/docs/5.2/layout/columns/#vertical-alignment>`_
    vertical_alignment: str | None = None

    def __init__(
        self,
        *columns: Block,
        horizontal_alignment: str | None = None,
        vertical_alignment: str | None = None,
        **kwargs: Any,
    ):
        self.horizontal_alignment = horizontal_alignment or self.horizontal_alignment
        self.vertical_alignment = vertical_alignment or self.vertical_alignment
        self.check_alignments()
        self.columns: list[Block] = list(columns)
        self.columns_map: dict[str, Block] = {}
        for i, column in enumerate(self.columns):
            name = column._name or f"column-{i + 1}"
            self.columns_map[name] = column
            self._add_helper_method(name)
        super().__init__(**kwargs)

    def check_alignments(self) -> None:
        """
        Check that our supplied ``horizontal_alignment`` and
        ``verttical_alignment`` settings to ensure they are valid values from
        :py:attr:`HORIZONTALALIGNMENTS` and :py:attr:`VERTICAL_ALIGNMENTS`,
        respectively.

        Raises:
            ValueError: an alignment was not valid

        """
        if (
            self.horizontal_alignment
            and self.horizontal_alignment not in self.HORIZONTAL_ALIGNMENT
        ):
            msg = (
                f'"{self.horizontal_alignment}" is not a valid horizontal alignment.  '
                f"Must be one of {', '.join(self.HORIZONTAL_ALIGNMENT)}"
            )
            raise ValueError(msg)
        if (
            self.vertical_alignment
            and self.vertical_alignment not in self.VERTICAL_ALIGNMENT
        ):
            msg = (
                f'"{self.vertical_alignment}" is not a valid vertical alignment.  '
                f"Must be one of {', '.join(self.VERTICAL_ALIGNMENT)}"
            )
            raise ValueError(msg)

    @property
    def column_names(self) -> list[str]:
        """
        Return the list of :py:attr:`Column.name` attributes of all of our
        columns.

        Returns:
            A list of column names.

        """
        return list(self.columns_map.keys())

    def _add_helper_method(self, name: str) -> None:
        """
        Add a method to this :py:class:`Row` object like so::

            def add_to_{column_name}(widget: Widget) -> None:
                ...

        This new method will allow you to add a widget to the column with name
        ``{column_name}`` directly without having to use
        :py:meth:`add_to_column`.

        Example:
            .. code-block:: python

                from wildewidgets.widgets import Row, Column, Block

                sidebar = Column(name='sidebar', width=3)
                main = Column(name='main')
                row = Row(columns=[sidebar, main])

            You can now add widgets to the sidebar column like so:

            .. code-block:: python

                widget = Block('foo')
                row.add_to_sidebar(widget)

        Args:
            name: the name of the column

        """
        name = re.sub("-", "_", name)
        setattr(self, f"add_to_{name}", partial(self.add_to_column, name))

    def add_column(self, column: Column) -> None:
        """
        Add a column to this row to the right of any existing columns.

        Note:
            A side effect of adding a column is to add a method to this
            :py:class:`Row` object like so:


            .. code-block:: python

                def add_to_{column_name}(widget: Widget) -> None:

            where ``{column_name}`` is either:

            * the value of ``column.name``, if that is not the default name
            * ``column-{n}`` where ``n`` is the number of columns in this row,
                starting with 1.

        Args:
            column: the column to add

        """
        if column._name:
            name: str = column._name
        else:
            name = f"column-{len(self.columns)}"
        self.add_block(column)
        self.columns.append(column)
        self.columns_map[name] = column
        self._add_helper_method(name)

    def add_to_column(self, identifier: int | str, block: Block) -> None:
        """
        Add ``widget`` to the column named ``identifier`` at the bottom of any
        other widgets in that column.

        Note:
            If ``identifier`` is an int, ``identifier`` should be 1-offset, not
            0-offset.

        Args:
            identifier: either a column number (left to right, starting with 1),
                or a column name
            block: the :py:class:`Block` subclass to append to this col

        """
        if isinstance(identifier, int):
            identifier = f"column-{identifier}"
        self.columns_map[identifier].add_block(block)


class TwoColumnLayout(Row):
    """
    A convenience widget that implements a two column ``row`` from the
    `Bootstrap Grid system <https://getbootstrap.com/docs/5.2/layout/grid/>`_
    with two named columns: ``left`` and ``right``.

    On viewports less than size ``md``, both columns will take up the entire width
    of the page (so that we look readable on portrait phones).

    Set the column widths for viewport ``md`` and above, define
    ``left_column_width`` as a class attribute or keyword argument, where 0 <
    ``left_column_width`` <= 12.  The default is to make the two columns
    equal widths.

    Examples:
        .. code-block:: python

            from wildewidgets import TwoColumnLayout, Block

            left_blocks = [Block('One', name='one'), Block('Two', name='two')]
            right_blocks = [Block('Three', name='three'), Block('Four', name='four')]
            layout = TwoColumnLayout(
                left_column_width=3,
                left_column_widgets=left_blocks,
                right_column_widgets=right_blocks
            )

    To create a two column layout with no blocks in either column:

        .. code-block:: python

            from wildewidgets import TwoColumnLayout

            layout = TwoColumnLayout(left_column_width=3)

    To create a two column layout with blocks in each column:

        .. code-block:: python

            from wildewidgets import TwoColumnLayout, Block

            layout = TwoColumnLayout(
                left_column_width=3,
                left_column_widgets=[
                    Block('One', name='one'), Block('Two', name='two')
                ],
                right_column_widgets=[
                    Block('Three', name='three'), Block('Four', name='four')
                ]
            )

    To add a block to a column after creating the ``TwoColumnLayout``:

        .. code-block:: python

            from wildewidgets import TwoColumnLayout, Block

            layout = TwoColumnLayout(left_column_width=3)
            layout.add_to_left_column(Block('One', name='one'))
            layout.add_to_right_column(Block('Two', name='two'))
            layout.add_to_left_column(Block('Three', name='three'))
            layout.add_to_right_column(Block('Four', name='four'))

    Keyword Args:
        left_column_width: the width of the left column, where 0 <
            ``left_column_width`` <= 12.  The right column will have its width set
            automatically based on this.
        left_column_widgets: A list of blocks to add to the left column
        right_column_widgets: A list of blocks to add to the right column

    """

    #: The name of this block
    name: str = "two-column"

    #: The width of the left column, where 0 < width <= 12.  The
    #: right column will have its width set automatically based on this.
    left_column_width: int = 6
    #: A list of blocks to add to the left column
    left_column_widgets: list[Block] = []  # noqa: RUF012
    #: A list of blocks to add to the right column
    right_column_widgets: list[Block] = []  # noqa: RUF012

    def __init__(
        self,
        left_column_width: int | None = None,
        left_column_widgets: list[Block] | None = None,
        right_column_widgets: list[Block] | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.left_column_width = left_column_width or self.left_column_width
        self.left_column_widgets = (
            left_column_widgets
            if left_column_widgets is not None
            else deepcopy(self.left_column_widgets)
        )
        self.right_column_widgets = (
            right_column_widgets
            if right_column_widgets is not None
            else deepcopy(self.right_column_widgets)
        )
        left_viewport_widths = {"md": str(self.left_column_width)}
        right_viewport_widths = {"md": str(12 - self.left_column_width)}
        self.add_column(
            Column(
                *self.left_column_widgets,
                name="left",
                base_width=12,
                viewport_widths=left_viewport_widths,
            )
        )
        self.add_column(
            Column(
                *self.right_column_widgets,
                name="right",
                base_width=12,
                viewport_widths=right_viewport_widths,
            )
        )



from __future__ import annotations

import math
import random
from typing import TYPE_CHECKING, Final, Literal

from django import template
from django.conf import settings

from wildewidgets.views import JSONDataView, WidgetInitKwargsMixin

from ..base import Widget

if TYPE_CHECKING:
    from collections.abc import Generator, Iterator


def float_range(
    start: float, stop: float, step: float = 1.0
) -> Generator[float, None, None]:
    """
    A generator function that yields a range of floating point numbers.

    This function behaves like the built-in `range` function, but for floating point
    numbers. It yields numbers starting from `start`, up to but not including `stop`,
    incrementing by `step`.

    Example:
        >>> list(float_range(0.0, 5.0, 1.0))
        [0.0, 1.0, 2.0, 3.0, 4.0]

        >>> list(float_range(0.0, 5.0, 0.5))
        [0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0]

    Note:
        This function does not support negative steps or reverse ranges.

    Args:
        start: the starting value of the range.
        stop: the end value of the range (exclusive).
        step: the increment between each value in the range.

    Yields:
        float: the next value in the range.

    """
    current = start
    while current < stop:
        yield current
        current += step


class CategoryChart(Widget, WidgetInitKwargsMixin, JSONDataView):
    """
    Base class for Chart.js charts that display data in categories.

    This abstract class provides the foundation for creating various Chart.js
    visualizations that organize data into categories (like bar charts, pie
    charts, etc.).  It handles dataset management, color assignment, and
    rendering.

    Note:
        This is an abstract base class. Subclasses should implement the necessary
        methods to create specific chart types.

    Example:
        .. code-block:: python

            from wildewidgets import CategoryChart

            class MyBarChart(CategoryChart):
                def __init__(self, **kwargs):
                    super().__init__(chart_type="bar", **kwargs)
                    self.set_categories(['A', 'B', 'C'])
                    self.add_dataset([10, 20, 30], "Series 1")

    """

    #: Default colors for the chart, used if no custom colors are set
    COLORS: Final[list[tuple[int, int, int]]] = [
        (0, 59, 76),
        (0, 88, 80),
        (100, 75, 120),
        (123, 48, 62),
        (133, 152, 148),
        (157, 174, 136),
        (159, 146, 94),
        (242, 211, 131),
        (30, 152, 138),
        (115, 169, 80),
    ]

    #: Default grayscale colors for the chart, used if no custom colors are set
    GRAYS: Final[list[tuple[int, int, int]]] = [
        (200, 200, 200),
        (229, 229, 229),
        (170, 169, 159),
        (118, 119, 123),
        (97, 98, 101),
        (175, 175, 175),
        (105, 107, 115),
    ]
    #: The Django template file to render the chart
    template_file: str = "wildewidgets/categorychart.html"
    #: Display legend for the chart, can be set in options
    legend: bool = False
    #: Position of the legend, can be set in options
    legend_position: Literal["top", "bottom", "left", "right"] = "top"
    #: Whether to use the color palette (True) or grayscale (False)
    color: bool = True

    def __init__(self, *args, **kwargs):
        self.chart_options = {
            "width": kwargs.get("width", "400px"),
            "height": kwargs.get("height", "400px"),
            "title": kwargs.get("title"),
            "legend": kwargs.get("legend", self.legend),
            "legend_position": kwargs.get("legend_position", self.legend_position),
            "chart_type": kwargs.get("chart_type"),
            "histogram": kwargs.get("histogram", False),
            "max": kwargs.get("max"),
            "thousands": kwargs.get("thousands", False),
            "histogram_max": kwargs.get("histogram_max"),
            "url": kwargs.get("url"),
        }
        self.chart_id = kwargs.get("chart_id")
        self.categories = None
        self.datasets = []
        self.dataset_labels = []
        self.color = kwargs.get("color", self.color)
        self.colors = []
        if hasattr(settings, "CHARTJS_FONT_FAMILY"):
            self.chart_options["chartjs_font_family"] = settings.CHARTJS_FONT_FAMILY
        if hasattr(settings, "CHARTJS_TITLE_FONT_SIZE"):
            self.chart_options["chartjs_title_font_size"] = (
                settings.CHARTJS_TITLE_FONT_SIZE
            )
        if hasattr(settings, "CHARTJS_TITLE_FONT_STYLE"):
            self.chart_options["chartjs_title_font_style"] = (
                settings.CHARTJS_TITLE_FONT_STYLE
            )
        if hasattr(settings, "CHARTJS_TITLE_PADDING"):
            self.chart_options["chartjs_title_padding"] = settings.CHARTJS_TITLE_PADDING
        super().__init__(*args, **kwargs)

    def set_categories(self, categories: list[str] | list[float]) -> None:
        """
        Set the categories for the chart's x-axis.

        Args:
            categories: List of category labels to display on the x-axis

        """
        self.categories = categories

    def add_dataset(
        self, dataset: list[float] | list[int], label: str | None = None
    ) -> None:
        """
        Add a dataset to the chart.

        Args:
            dataset: List of data values corresponding to each category
            label: Optional name for the dataset (shown in legend)

        """
        self.datasets.append(dataset)
        self.dataset_labels.append(label)

    def set_option(self, name: str, value: str | float | bool | None) -> None:
        """
        Set a chart configuration option.

        Args:
            name: The option name
            value: The option value

        """
        self.chart_options[name] = value

    def set_color(self, color: bool) -> None:
        """
        Enable or disable color mode.

        Args:
            color: If True, use the color palette; if False, use grayscale

        """
        self.color = color

    def set_colors(self, colors: list[tuple[int, int, int]]) -> None:
        """
        Set a custom color palette for the chart.

        Args:
            colors: List of RGB color tuples to use for chart elements

        """
        self.colors = colors

    def get_content(self, **kwargs):  # noqa: ARG002
        """
        Render the chart as HTML content.

        Returns:
            str: The rendered HTML for the chart

        """
        chart_id = self.chart_id or str(random.randrange(0, 1000))  # noqa: S311
        template_file = self.template_file
        context = self.get_context_data() if self.datasets else {"async": True}
        html_template = template.loader.get_template(template_file)
        context["options"] = self.chart_options
        context["name"] = f"chart_{chart_id}"
        context["wildewidgetclass"] = self.__class__.__name__
        context["extra_data"] = self.get_encoded_extra_data()
        return html_template.render(context)

    def __str__(self):
        """
        Return the string representation of the chart.

        Returns:
            str: The rendered HTML for the chart

        """
        return self.get_content()

    def get_context_data(self, **kwargs):
        """
        Get the context data for rendering the chart template.

        Returns:
            dict: The context data for the template

        """
        context = super().get_context_data(**kwargs)
        self.load()
        context.update(
            {
                "labels": self.get_categories(),
                "datasets": self.get_dataset_configs(),
            }
        )
        return context

    def get_color_iterator(self) -> Iterator[tuple[int, int, int]]:
        """
        Get an iterator over the chart's color palette.

        Returns:
            iterator: Iterator over RGB color tuples

        """
        if self.colors:
            return iter(self.colors)
        if self.color:
            return iter(self.COLORS)
        return iter(self.GRAYS)

    def get_dataset_options(self, index, color: tuple[int, int, int]):  # noqa: ARG002
        """
        Get rendering options for a dataset.

        Args:
            index: The index of the dataset
            color: The RGB color tuple for the dataset

        Returns:
            dict: Dataset rendering options

        """
        return {
            "backgroundColor": f"rgba({color[0]}, {color[1]}, {color[2]}, 0.5)",
            "borderColor": f"rgba({color[0]}, {color[1]}, {color[2]}, 1)",
            "borderWidth": 0.2,
        }

    def get_dataset_configs(self) -> list[dict]:
        """
        Get the complete configuration for all datasets.

        Returns:
            list[dict]: List of dataset configuration dictionaries

        Note:
            This method should be implemented by subclasses.

        """
        return []

    def get_categories(self) -> list[str]:
        """
        Get the category labels for the chart.

        Returns:
            list[str]: List of category labels

        Raises:
            NotImplementedError: If not implemented by a subclass

        """
        if self.categories:
            return self.categories
        msg = 'You should return a labels list. (i.e: ["January", ...])'
        raise NotImplementedError(msg)

    def get_datasets(self) -> list[list[float]]:
        """
        Get all datasets for the chart.

        Returns:
            list[list[float]]: List of datasets, each containing values for each
                category

        Raises:
            NotImplementedError: If not implemented by a subclass

        """
        if self.datasets:
            return self.datasets
        msg = "You should return a data list list. (i.e: [[25, 34, 0, 1, 50], ...])."
        raise NotImplementedError(msg)

    def get_dataset_labels(self) -> list[str]:
        """
        Get labels for all datasets.

        Returns:
            list[str]: List of dataset labels

        """
        return self.dataset_labels

    def load(self) -> None:
        """
        Load data into the chart.

        This method should be implemented by subclasses to load data from
        external sources or perform calculations before rendering.
        """


class DoughnutChart(CategoryChart):
    """
    A doughnut chart implementation using Chart.js.

    This class creates a doughnut chart (a pie chart with a hole in the center)
    that displays data as segments of a circle. Each segment represents a
    proportion of the whole, making this chart ideal for showing composition.

    Example:
        .. code-block:: python

            chart = DoughnutChart(title="Browser Usage")
            chart.set_categories(['Chrome', 'Firefox', 'Safari', 'Edge', 'Other'])
            chart.add_dataset([65, 15, 10, 8, 2], "Usage Share")

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize a doughnut chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class
                chart_type: Set to "doughnut" by default if not specified

        """
        if "chart_type" not in kwargs:
            kwargs["chart_type"] = "doughnut"
        super().__init__(*args, **kwargs)

    def get_dataset_configs(self) -> list[dict]:
        """
        Configure datasets for the doughnut chart.

        For doughnut charts, each data point gets its own color in the
        backgroundColor array.

        Returns:
            list[dict]: List containing a single dataset configuration

        """
        datasets = []
        color_generator = self.get_color_iterator()
        data = self.get_dataset()
        dataset = {"data": data}
        dataset["backgroundColor"] = []
        for _ in range(len(data)):
            color = tuple(next(color_generator))
            dataset["backgroundColor"].append(
                f"rgba({color[0]}, {color[1]}, {color[2]}, 0.5)"
            )
        datasets.append(dataset)
        return datasets

    def get_dataset(self):
        """
        Get the primary dataset for the doughnut chart.

        Returns:
            list: The first dataset in the datasets list

        """
        return self.datasets[0]


class PieChart(DoughnutChart):
    """
    A pie chart implementation using Chart.js.

    This class creates a pie chart that displays data as segments of a circle.
    Each segment represents a proportion of the whole, making this chart ideal
    for showing composition.

    It extends DoughnutChart and has the same functionality, just with a different
    chart type.

    Example:
        .. code-block:: python

            from wildewidgets import PieChart

            chart = PieChart(title="Expenditure Breakdown")
            chart.set_categories(
                ['Housing', 'Food', 'Transport', 'Entertainment', 'Other']
            )
            chart.add_dataset([35, 25, 15, 15, 10], "Percentage")

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a pie chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class

        Note:
            Automatically sets chart_type to "pie"

        """
        kwargs["chart_type"] = "pie"
        super().__init__(*args, **kwargs)


class BarChart(CategoryChart):
    """
    A bar chart implementation using Chart.js.

    This class creates a vertical bar chart where data is displayed as rectangular
    bars with heights proportional to their values. Bar charts are excellent for
    comparing quantities across different categories.

    Features:

    - Support for multiple datasets (grouped bars)
    - Optional stacking of bars
    - Horizontal orientation option
    - Money formatting option

    Example:
        .. code-block:: python

            from wildewidgets import BarChart

            chart = BarChart(title="Monthly Sales")
            chart.set_categories(['Jan', 'Feb', 'Mar', 'Apr', 'May'])
            chart.add_dataset([10000, 15000, 12000, 18000, 20000], "2022")
            chart.add_dataset([12000, 18000, 15000, 21000, 25000], "2023")

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a bar chart.

        Args:
            *args: Positional arguments passed to the parent class

        Keyword Args:
            **kwargs: Keyword arguments passed to the parent class
                chart_type: Set to "bar" by default if not specified
                money: Whether to format values as currency (default: False)
                stacked: Whether to stack multiple datasets (default: False)
                xAxes_name: Name for x-axis configuration (default: "xAxes")
                yAxes_name: Name for y-axis configuration (default: "yAxes")

        """
        if "chart_type" not in kwargs:
            kwargs["chart_type"] = "bar"
        super().__init__(*args, **kwargs)
        self.set_option("money", kwargs.get("money", False))
        self.set_option("stacked", kwargs.get("stacked", False))
        self.set_option("xAxes_name", kwargs.get("xAxes_name", "xAxes"))
        self.set_option("yAxes_name", kwargs.get("yAxes_name", "yAxes"))

    def set_stacked(self, stacked: bool) -> None:
        """
        Enable or disable stacked mode for the bar chart.

        In stacked mode, bars from different datasets are stacked on top of
        each other instead of being displayed side by side.

        Args:
            stacked: If True, enable stacked mode; if False, disable it

        """
        self.set_option("stacked", "true" if stacked else "false")

    def set_horizontal(self, horizontal: bool) -> None:
        """
        Set the orientation of the bar chart.

        Args:
            horizontal: If True, display bars horizontally; if False, display vertically

        """
        if horizontal:
            self.chart_options["xAxes_name"] = "yAxes"
            self.chart_options["yAxes_name"] = "xAxes"
            self.chart_options["chart_type"] = "horizontalBar"
        else:
            self.chart_options["xAxes_name"] = "xAxes"
            self.chart_options["yAxes_name"] = "yAxes"
            self.chart_options["chart_type"] = "bar"

    def get_dataset_options(self, index, color: tuple[int, int, int]):  # noqa: ARG002
        """
        Get rendering options for a dataset in the bar chart.

        Args:
            index: The index of the dataset
            color: The RGB color tuple for the dataset

        Returns:
            dict: Dataset rendering options with appropriate styling

        """
        default_opt: dict[str, str | float] = {
            "backgroundColor": f"rgba({color[0]}, {color[1]}, {color[2]}, 0.65)",
        }
        if not self.chart_options["histogram"]:
            default_opt["borderColor"] = f"rgba({color[0]}, {color[1]}, {color[2]}, 1)"
            default_opt["borderWidth"] = 0.2
        return default_opt

    def get_dataset_configs(self) -> list[dict]:
        """
        Get the complete configuration for all datasets in the bar chart.

        This method prepares all datasets with appropriate colors and options.

        Returns:
            list[dict]: List of dataset configuration dictionaries

        """
        datasets = []
        color_generator = self.get_color_iterator()
        data = self.get_datasets()
        dataset_labels = self.get_dataset_labels()
        num = len(dataset_labels)
        for i, entry in enumerate(data):
            color = next(color_generator)
            dataset: dict[str, str | float | list[float]] = {"data": entry}
            dataset.update(self.get_dataset_options(i, color))
            if i < num:
                dataset["label"] = dataset_labels[i]  # series labels for Chart.js
                dataset["name"] = dataset_labels[i]  # HighCharts may need this
            datasets.append(dataset)
        return datasets


class StackedBarChart(BarChart):
    """
    A stacked bar chart implementation using Chart.js.

    This class creates a vertical bar chart where multiple datasets are stacked
    on top of each other for each category. This is useful for comparing total
    values across categories while also showing the composition of each total.

    Example:
        .. code-block:: python

            from wildewidgets import StackedBarChart

            chart = StackedBarChart(title="Revenue Breakdown")
            chart.set_categories(['Q1', 'Q2', 'Q3', 'Q4'])
            chart.add_dataset([50000, 60000, 55000, 75000], "Product A")
            chart.add_dataset([35000, 40000, 45000, 55000], "Product B")
            chart.add_dataset([15000, 20000, 25000, 30000], "Product C")

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize a stacked bar chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class

        Note:
            Automatically enables stacked mode during initialization.

        """
        super().__init__(*args, **kwargs)
        self.set_stacked(True)


class HorizontalBarChart(BarChart):
    """
    A horizontal bar chart implementation using Chart.js.

    This class creates a horizontal bar chart where data is displayed as
    rectangular bars with lengths proportional to their values. Horizontal
    bar charts are excellent for comparing quantities across different
    categories, especially when category labels are long.

    Example:
        .. code-block:: python

            from wildewidgets import HorizontalBarChart

            chart = HorizontalBarChart(title="Population by Country")
            chart.set_categories(
               ['United States', 'Indonesia', 'Brazil', 'Russia', 'Mexico']
            )
            chart.add_dataset(
                [331000000, 273500000, 211800000, 145500000, 130000000], "Population"
            )

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize a horizontal bar chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class

        Note:
            Automatically sets horizontal orientation during initialization.

        """
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)


class HorizontalStackedBarChart(BarChart):
    """
    A horizontal stacked bar chart implementation using Chart.js.

    This class creates a horizontal bar chart where multiple datasets are stacked
    on top of each other for each category. Useful for comparing parts of a whole
    across different categories.

    The chart is configured with both horizontal orientation and stacked mode
    automatically during initialization.

    Examples:
        .. code-block:: python

            from wildewidgets import HorizontalStackedBarChart

            chart = HorizontalStackedBarChart()
            chart.set_categories(['A', 'B', 'C'])
            chart.add_dataset([10, 20, 30], "Dataset 1")
            chart.add_dataset([5, 15, 25], "Dataset 2")

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize a horizontal stacked bar chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class

        Note:
            Automatically configures the chart as both horizontal and stacked.

        """
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)
        self.set_stacked(True)


class Histogram(BarChart):
    """
    A vertical histogram chart for visualizing data distribution.

    This class creates a histogram that displays the distribution of numerical data
    by grouping values into bins. The height of each bar represents the frequency
    of values within that bin.

    To use this chart, call the `build` method with your raw data and the desired
    number of bins. The class will automatically:
    1. Calculate appropriate bin ranges
    2. Count values in each bin
    3. Configure the chart with proper categories and dataset

    Examples:
        .. code-block:: python

            # Create a histogram with 10 bins for a list of data points
            chart = Histogram(title="Distribution of Values")
            chart.build([23.5, 24.1, 25.3, 26.2, 27.5, 28.1, 29.3, 30.2, 31.5], 10)

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize a histogram chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class
                histogram: Flag to enable histogram mode (set to True by default)

        """
        if "histogram" not in kwargs:
            kwargs["histogram"] = True
        super().__init__(*args, **kwargs)

    def build(self, data: list[float], bin_count: int) -> None:
        """
        Process raw data and configure the histogram.

        This method handles all the calculations needed to convert raw data into
        a histogram representation:
        - Determines the range of the data
        - Calculates appropriate bin sizes and boundaries
        - Counts the frequency of values in each bin
        - Sets up the chart categories and dataset

        The method uses a robust algorithm that handles both positive and negative
        values, and attempts to create visually appealing bin boundaries.

        Args:
            data: List of numerical values to be plotted in the histogram
            bin_count: Number of bins to divide the data into

        Note:
            The method will set two special options on the chart:
            - "max": The upper bound of the last visible bin
            - "histogram_max": The absolute upper boundary of the data range

        """
        num_min = min(data)
        num_max = max(data)

        num_range = num_max - num_min
        bin_chunk = num_range / bin_count
        bin_power = math.floor(math.log10(bin_chunk))
        bin_chunk = math.ceil(bin_chunk / 10**bin_power) * 10**bin_power
        if num_min < 0:
            bin_min = math.ceil(math.fabs(num_min / bin_chunk)) * bin_chunk * -1
        else:
            bin_min = math.floor(num_min / bin_chunk) * bin_chunk
        if num_max < 0:
            bin_max = math.floor(math.fabs(num_max / bin_chunk)) * bin_chunk * -1
        else:
            bin_max = math.ceil(num_max / bin_chunk) * bin_chunk
        categories = list(float_range(bin_min, bin_max + bin_chunk, bin_chunk))
        self.set_option("max", categories[-2])
        self.set_option("histogram_max", categories[-1])
        bins = [0] * bin_count
        for num in data:
            for i in range(len(categories) - 1):
                if num >= categories[i] and num < categories[i + 1]:
                    if i < len(bins):
                        bins[i] += 1
        self.set_categories(categories)
        self.add_dataset(bins, "data")


class HorizontalHistogram(Histogram):
    """
    A horizontal histogram chart for visualizing data distribution.

    This class extends the regular Histogram to display bars horizontally instead
    of vertically. This can be particularly useful when:
    - You have many bins that would appear too narrow in a vertical histogram
    - You want to display long category labels that are easier to read horizontally
    - You want to emphasize the comparison of frequencies across bins

    All the functionality of the regular Histogram is preserved, including the
    automatic calculation of bin ranges and frequencies.

    Examples:
        .. code-block:: python

            # Create a horizontal histogram with 8 bins
            chart = HorizontalHistogram(title="Age Distribution")
            chart.build([21, 22, 24, 25, 26, 28, 29, 31, 32, 33, 35, 37, 39, 45], 8)

    """

    def __init__(self, *args, **kwargs):
        """
        Initialize a horizontal histogram chart.

        Args:
            *args: Positional arguments passed to the parent class
            **kwargs: Keyword arguments passed to the parent class

        Note:
            Automatically configures the chart with horizontal orientation while
            preserving all histogram functionality.

        """
        super().__init__(*args, **kwargs)
        self.set_horizontal(True)



from __future__ import annotations

from copy import copy
from dataclasses import dataclass
from typing import Any

from .base import Block, Widget, WidgetStream
from .buttons import FormButton, LinkButton
from .headers import PageHeader, WidgetListLayoutHeader


@dataclass
class WidgetIndexItem:
    """
    Data container for items in a widget index.

    This dataclass represents a single entry in a widget index, combining
    a widget instance with its display title and icon.

    Attributes:
        widget: The widget instance to be displayed in the index
        title: Display title for the widget in the index
        icon: Icon identifier to display next to the widget title, defaults to "gear"

    """

    widget: Widget
    title: str
    icon: str = "gear"


class WidgetIndex(Block):
    """
    A navigational index of widgets.

    This widget creates a visual index or table of contents for a collection
    of widgets. Each entry in the index consists of a widget with an associated
    title and icon.

    Note:
        This is not something that you would use directly.
        :py:class:`WidgetListLayout` uses this internally when you add widgets
        to the main content area.

    Attributes:
        template_name: Path to the template used to render the widget index
        block: CSS block name for styling
        entries: Default list of widget index items

    Args:
        *args: Positional arguments passed to parent

    Keyword Args:
        **kwargs: Keyword arguments which may include 'entries' to
            initialize the index with a list of :py:class:`WidgetIndexItem` objects


    """

    template_name: str = "wildewidgets/widget_index.html"
    block: str = "widget-index"
    entries: list[WidgetIndexItem] = []  # noqa: RUF012

    def __init__(self, *args, **kwargs):
        """
        Initialize a widget index.

        """
        entries = kwargs.pop("entries", copy(self.entries))
        super().__init__(*args, **kwargs)
        self._entries = entries

    @property
    def is_empty(self) -> bool:
        """
        Check if the widget index is empty.

        Returns:
            bool: True if the index contains no entries, False otherwise

        """
        return len(self._entries) == 0

    def add_widget(
        self, widget: Widget, title: str | None = None, icon: str | None = None
    ) -> None:
        """
        Add a widget to the index.

        Adds the given widget to the index with an associated title and icon.
        If title or icon are not provided, they will be determined from the
        widget attributes or defaults.

        Note:
            If the widget has an is_visible function that evaluates to False,
            the widget will not be added to the index.

        Args:
            widget: The widget to add to the index
            title: Optional display title for the widget. If None, uses `widget.title1`
                  or the widget class name
            icon: Optional icon identifier. If None, uses widget.icon or "gear"

        """
        if hasattr(widget, "is_visible") and not widget.is_visible():
            return
        item = WidgetIndexItem(
            widget=widget,
            title=getattr(widget, "title", widget.__class__.__name__),
            icon=getattr(widget, "icon", "gear"),
        )
        if title is not None:
            item.title = title
        if icon is not None:
            item.icon = icon
        self._entries.append(item)

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        """
        Get context data for template rendering.

        Adds the widget index entries to the template context.

        Args:
            *args: Positional arguments passed to parent
            **kwargs: Additional context variables

        Returns:
            dict: Updated context dictionary with entries

        """
        context = super().get_context_data(*args, **kwargs)
        context["entries"] = self._entries
        return context


class WidgetListSidebarWidget(Block):
    """
    A sidebar component for widget list layouts.

    This widget creates a sidebar that can contain actions (typically buttons)
    and other widgets. It's commonly used as part of a WidgetListLayout to
    provide navigation and action controls.

    Note:
        This is not something that you would use directly.
        :py:class:`WidgetListLayout` uses this internally when you add widgets
        to the sidebar area.

    Attributes:
        template_name: Path to the template used to render the sidebar
        block: CSS block name for styling
        css_class: CSS classes to apply to the sidebar
        actions: Default list of action widgets to display
        bare_widgets: Default list of widgets to display without special styling

    Args:
        *args: Positional arguments passed to parent

    Keyword Args:
        title: Optional title for the sidebar
        width: Column width for the sidebar, defaults to 3
        breakpoint: Bootstrap breakpoint for responsive design
        bare_widgets: Optional list of widgets to display without special styling
        actions: Optional list of action widgets to display
        **kwargs: Additional keyword arguments passed to parent

    """

    template_name: str = "wildewidgets/widget-list--sidebar.html"

    block: str = "widget-list__sidebar"
    css_class: str = ""
    actions: list[Widget] = []  # noqa: RUF012
    bare_widgets: list[Widget] = []  # noqa: RUF012

    class Widgets(WidgetStream):
        block: str = "widget-list__sidebar__widgets"

    class Actions(WidgetStream):
        css_class: str = (
            "px-3 py-4 d-flex flex-column align-items-stretch border bg-white shadow-sm"
        )

    def __init__(
        self,
        *args,
        title: str | None = None,
        width: int = 3,
        breakpoint: str = "xl",  # noqa: A002
        bare_widgets: list[Widget] | None = None,
        actions: list[Widget] | None = None,
        **kwargs,
    ):
        if title is not None:
            self.title = title
        if self.css_class is None:
            self.css_class = ""
        if breakpoint:
            self.css_class += f"col-{breakpoint}-{width}"
        else:
            self.css_class += f"col-{width}"
        super().__init__(*args, **kwargs)
        actions = actions if actions is not None else self.actions
        bare_widgets = bare_widgets if bare_widgets is not None else self.bare_widgets
        self.widget_index = WidgetIndex()
        self._widgets = WidgetListSidebarWidget.Widgets(widgets=bare_widgets)
        self._actions = WidgetListSidebarWidget.Actions(widgets=actions)
        self._actions.block = f"{self.block}__actions"

    def add_link_button(self, text: str, url: str, **kwargs) -> None:
        """
        Add :class:`wildewidgets.LinkButton` to the sidebar.

        Args:
            text: The text to display on the button.
            url: The URL to link to.

        Keyword Args:
            **kwargs: any additional keyword arguments appropriate for
                      :class:`wildewidgets.FormButton`

        """
        kwargs["text"] = text
        kwargs["url"] = url
        if "css_class" in kwargs:
            kwargs["css_class"] = f"{kwargs['css_class']} w-100"
        else:
            kwargs["css_class"] = "w-100"
        self.add_actions_widget(LinkButton(**kwargs))

    def add_form_button(self, text: str, action: str, **kwargs) -> None:
        """
        Add :class:`wildewidgets.FormButton` to the sidebar.

        Args:
            text: The text to display on the button.
            action: The action to perform when the button is clicked.

        Keyword Args:
            **kwargs: any additional keyword arguments appropriate for
                      :class:`wildewidgets.FormButton`

        """
        kwargs["text"] = text
        kwargs["action"] = action
        if "css_class" in kwargs:
            kwargs["css_class"] = f"{kwargs['css_class']} w-100"
        else:
            kwargs["css_class"] = "w-100"
        if "button_css_class" in kwargs:
            kwargs["button_css_class"] = f"{kwargs['button_css_class']} w-100"
        else:
            kwargs["button_css_class"] = "w-100"
        self.add_actions_widget(FormButton(**kwargs))

    def add_widget(self, widget: Widget) -> None:
        """
        Add a widget to the sidebar outside the Actions box.
        """
        self._widgets.add_widget(widget)

    def add_actions_widget(self, widget: Widget) -> None:
        """
        Add a widget to the sidebar inside the Actions box.
        """
        self._actions.add_widget(widget)

    def add_widget_to_index(
        self, widget: Widget, title: str | None = None, icon: str | None = None
    ) -> None:
        """
        Add ``widget`` the index with title ``title`` and icon ``icon``.

        If ``title`` is ``None``, look for a title on ``widget.title``.  If that
        is also ``None``, default to the name of the widget class.

        If ``icon`` is ``None``, look for an icon on ``widget.icon``.  If that
        is also ``None``, default to the Bootstrap Icons "gear" icon.
        """
        self.widget_index.add_widget(widget, title=title, icon=icon)

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(*args, **kwargs)
        context["title"] = self.title
        if not self._widgets.is_empty:
            context["widgets"] = self._widgets
            self.widget_index._css_class = "mt-5"
        if not self._actions.is_empty:
            context["actions"] = self._actions
            self.widget_index._css_class = "mt-5"
        if not self.widget_index.is_empty:
            context["widget_index"] = self.widget_index
        return context


class WidgetListMainWidget(Block):
    """
    The main content area in a widget list layout.

    This widget represents the primary content area where widgets are displayed
    in a vertical stack. It's typically used in conjunction with a sidebar
    widget to create a comprehensive layout.

    Note:
        This is not something that you would use directly.
        :py:class:`WidgetListLayout` uses this internally when you add widgets
        to the main area.

    Attributes:
        template_name: Path to the template used to render the main widget list
        block: CSS block name for styling
        css_class: CSS class for the main content area
        entry_css_class: CSS class for individual widget entries
        entry_title_css_class: CSS class for widget entry titles
        entries: Default list of widget index items for the main content

    Args:
        *args: Positional arguments passed to parent

    Keyword Args:
        **kwargs: Keyword arguments which may include 'entries' to
            initialize the list with a list of WidgetIndexItem objects


    """

    template_name: str = "wildewidgets/widget-list--main.html"
    block: str = "widget-list__main"

    css_class: str = "col"
    entry_css_class: str | None = "shadow bg-white"
    entry_title_css_class: str | None = "font-weight-bold"
    entries: list[Widget] = []  # noqa: RUF012

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize the main widget list.

        """
        entries = kwargs.pop("entries", copy(self.entries))
        self._entry_css_class = kwargs.pop("entry_css_class", self.entry_css_class)
        self._entry_title_css_class = kwargs.pop(
            "entry_title_css_class", self.entry_title_css_class
        )
        super().__init__(*args, **kwargs)
        self._entries: list[Widget] = entries

    def add_widget(self, widget: Widget, title: str | Widget | None = None) -> None:
        """
        Add a widget to the main content area.

        This method also ensures that the widget has an associated title,
        either from the widget itself or as a provided argument.

        Args:
            widget: The widget to add to the main content area
            title: Optional title for the widget. If provided, it will be
                   set as the widget's title attribute

        """
        if hasattr(widget, "is_visible") and not widget.is_visible():
            return
        if title is not None:
            widget.title = title
        widget_title = widget.get_title()
        if not isinstance(widget_title, Widget):
            header = WidgetListLayoutHeader(header_text=widget.title)
            widget.title = header
        self._entries.append(widget)

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        """
        Get context data for template rendering.

        Provides the list of entries to be rendered in the main widget area.

        Args:
            *args: Positional arguments passed to parent
            **kwargs: Additional context variables

        Returns:
            dict: Updated context dictionary with entries and CSS classes

        """
        context = super().get_context_data(*args, **kwargs)
        context["entries"] = self._entries
        context["entry_title_css_class"] = self._entry_title_css_class
        context["entry_css_class"] = self._entry_css_class
        return context


class WidgetListLayout(Block):
    """
    Provides a two column layout. The first column is the sidebar containing
    links to the various widgets in the right column, and the second column is
    the main content, consisting of the contained widgets vertically stacked.

    This is really the only class you need to use to create a page with a
    sidebar and a main content area.  You can add widgets to the sidebar and
    the main content area, and they will be displayed in the appropriate places.

    Example:
        .. code-block:: python

            from wildewidgets import WidgetListLayout
            from django.urls import reverse
            from core.widgets import Widget1, Widget2, Widget3

            layout = WidgetListLayout('My Page')
            layout.add_sidebar_form_button('Update', reverse('core:thing--update'))
            layout.add_widget(Widget1(), title='another title')
            layout.add_widget(Widget2(), title='the title', icon='stuff')
            layout.add_widget(Widget3())

    Keyword Args:
        title: The title of the widget.
        sidebar_title: the title of the sidebar column
        sidebar_width: the width in columns of the sidebar column
        sidebar_breakpoint: the breakpoint at which the sidebar will collapse
            to a hamburger menu.  Defaults to "xl" (extra large).
        **kwargs: Additional keyword arguments passed to the parent
            *:py:class:`wildewidgets.Block` class

    """

    #: The Django template to render this layout.
    template_name: str = "wildewidgets/widget-list.html"
    #: The title of the sidebar column.
    sidebar_title: str = "Actions"
    #: The width of the sidebar column in Bootstrap grid columns.
    sidebar_width: int = 3
    #: The breakpoint at which the sidebar will collapse to a hamburger menu.
    sidebar_breakpoint: str = "xl"

    def __init__(
        self,
        title: str,
        sidebar_title: str | None = None,
        sidebar_width: int | None = None,
        sidebar_breakpoint: str | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.title = title
        self.header: Widget = PageHeader(header_text=title)
        self.sidebar: Widget = WidgetListSidebarWidget(
            title=sidebar_title if sidebar_title is not None else self.sidebar_title,
            width=sidebar_width if sidebar_width is not None else self.sidebar_width,
            breakpoint=sidebar_breakpoint
            if sidebar_breakpoint is not None
            else self.sidebar_breakpoint,
        )
        self.main = WidgetListMainWidget()
        self.modals: list[Widget] = []

    def add_widget(
        self, widget: Widget, title: str | None = None, icon: str | None = None
    ) -> None:
        """
        Add a widget to the layout.

        This method adds the widget to both the sidebar index and the main
        content area.

        Args:
            widget: The widget to add
            title: Optional title for the widget
            icon: Optional icon for the widget

        """
        self.sidebar.add_widget_to_index(widget, title=title, icon=icon)
        self.main.add_widget(widget, title=title)

    def add_modal(self, modal: Widget) -> None:
        """
        Add a modal widget to the layout.

        Modals are typically used for dialogs or secondary content that
        overlays the main content.

        Args:
            modal: The modal widget to add

        Returns:
            None

        """
        self.modals.append(modal)

    def add_sidebar_link_button(self, text: str, url: str, **kwargs) -> None:
        """
        Add a link button to the sidebar.

        This method simplifies the addition of link buttons to the sidebar
        with appropriate styling and behavior.

        Args:
            text: The text to display on the button
            url: The URL that the button should link to
            **kwargs: Additional keyword arguments for customization

        """
        self.sidebar.add_link_button(text, url, **kwargs)

    def add_sidebar_form_button(self, text: str, action: str, **kwargs) -> None:
        """
        Add a form button to the sidebar.

        This method simplifies the addition of form buttons to the sidebar
        with appropriate styling and behavior.

        Args:
            text: The text to display on the button
            action: The form action URL that the button should submit to
            **kwargs: Additional keyword arguments for customization

        """
        self.sidebar.add_form_button(text, action, **kwargs)

    def add_sidebar_widget(self, widget: Widget) -> None:
        """
        Add a widget to the sidebar.

        This method adds the widget to the Actions area of the sidebar,
        typically used for secondary actions or information.

        Args:
            widget: The widget to add to the sidebar

        """
        self.sidebar.add_actions_widget(widget)

    def add_sidebar_bare_widget(self, widget: Widget) -> None:
        """
        Add a bare widget to the sidebar.

        Bare widgets are added outside the Actions area and are not
        subject to special styling.

        Args:
            widget: The widget to add to the sidebar

        """
        self.sidebar.add_widget(widget)

    def get_context_data(self, *args, **kwargs) -> dict[str, Any]:
        """
        Get context data for template rendering.

        Provides the layout with the necessary context variables for
        rendering the title, sidebar, main content, and any modals.

        Args:
            *args: Positional arguments passed to parent
            **kwargs: Additional context variables

        Returns:
            dict: Updated context dictionary with layout components

        """
        context = super().get_context_data(*args, **kwargs)
        context["title"] = self.title
        context["sidebar"] = self.sidebar
        context["sidebar_width"] = self.sidebar_width
        context["main"] = self.main
        context["modals"] = self.modals
        context["header"] = self.header
        return context



from __future__ import annotations

import inspect
import random
from typing import TYPE_CHECKING, ClassVar, Literal

import altair as alt
import pandas as pd
from book_manager.forms import MinimalBookForm
from book_manager.models import Author, Book
from wildewidgets import (
    AltairChart,
    ApexChartBase,
    ApexDatasetBase,
    ApexJSONMixin,
    ApexSparkline,
    BasicHeader,
    BasicModelTable,
    Block,
    CardHeader,
    CardWidget,
    CodeWidget,
    CollapseWidget,
    CrispyFormModalWidget,
    CrispyFormWidget,
    DataTable,
    DataTableFilter,
    DoughnutChart,
    FontIcon,
    Histogram,
    HorizontalHistogram,
    HorizontalLayoutBlock,
    HorizontalStackedBarChart,
    HTMLWidget,
    LabelBlock,
    LinkButton,
    ListModelCardWidget,
    ListModelWidget,
    MarkdownWidget,
    ModalButton,
    ModalWidget,
    PagedModelWidget,
    PageHeader,
    PieChart,
    StackedBarChart,
    StaticTableWidget,
    TabbedWidget,
    TablerFontIcon,
    TagBlock,
    TimeStamp,
    UnorderedList,
    Widget,
    WidgetCellMixin,
    WidgetStream,
)

from .models import Measurement

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet


def get_code_block(fn):
    """
    The lines are part of a function, so we need to remove the
    first line, which contains the function name, and the last
    line, which is just the return, and we need to remove any
    indentation on the remaining lines.

    Args:
        fn: the function to inspect

    """
    lines = inspect.getsourcelines(fn)[0]
    code = ""
    if len(lines) < 2:  # noqa: PLR2004
        return None
    space_len = len(lines[1]) - len(lines[1].strip())
    for line in lines[1:-1]:
        code += line[space_len - 1 :]
    return code


class BusinessChartHeader(BasicHeader):
    """
    Basic business charts with ChartJS.
    """

    header_level = 5
    header_type = "display"
    header_text = "Basic Business Charts w/ ChartJS"


class TestChart(StackedBarChart):
    """
    Test chart with ChartJS.
    """

    def get_categories(self) -> list[str]:
        """
        Return 7 labels for the x-axis.

        Returns:
            A list of 7 labels for the x-axis.

        """
        return ["January", "February", "March", "April", "May", "June", "July"]

    def get_dataset_labels(self) -> list[str]:
        """
        Return names of datasets.

        Returns:
            A list of names of datasets.

        """
        return [
            "Central",
            "Eastside",
            "Westside",
            "Central2",
            "Eastside2",
            "Westside2",
            "Central3",
            "Eastside3",
            "Westside3",
        ]

    def get_datasets(self) -> list[list[float]]:
        """Return 3 datasets to plot."""
        return [
            [750, 440, 920, 1100, 440, 950, 350],
            [410, 1920, 180, 300, 730, 870, 920],
            [870, 210, 940, 3000, 900, 130, 650],
            [750, 440, 920, 1100, 440, 950, 350],
            [410, 920, 180, 2000, 730, 870, 920],
            [870, 210, 940, 300, 900, 130, 650],
            [750, 440, 920, 1100, 440, 950, 3500],
            [410, 920, 180, 3000, 730, 870, 920],
            [870, 210, 940, 300, 900, 130, 650],
        ]


class TestDoughnutChart(DoughnutChart):
    """
    Test doughnut chart with ChartJS.
    """

    #: Whether to show the legend.
    legend = True
    #: The position of the legend.
    legend_position = "right"
    #: Whether to color the chart.
    color = False

    def get_categories(self) -> list[str]:
        """
        Return names of categories.

        Returns:
            A list of names of categories.

        """
        return ["January", "February", "March", "April", "May", "June", "July"]

    def get_dataset(self) -> list[int]:
        """
        Return a dataset.

        Returns:
            A list of values for the dataset.

        """
        return [75, 44, 92, 11, 44, 95, 53]


class TestHistogram(Histogram):
    """
    Test histogram with ChartJS.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        mu = 0
        sigma = 50
        nums = []
        bin_count = 40
        for _ in range(10000):
            temp = random.gauss(mu, sigma)
            nums.append(temp)

        self.build(nums, bin_count)


class TestHorizontalHistogram(HorizontalHistogram):
    """
    Test horizontal histogram with ChartJS.
    """

    #: Whether to color the chart.
    color = False

    def load(self) -> None:
        """
        Load the histogram data.

        """
        mu = 100
        sigma = 30
        nums = []
        bin_count = 50
        for _ in range(10000):
            temp = random.gauss(mu, sigma)
            nums.append(temp)

        self.build(nums, bin_count)


class SciChartHeader(BasicHeader):
    """
    Header for the scientific charts with Altair page.
    """

    #: The level of the header.
    header_level = 5
    #: The type of the header.
    header_type = "display"
    #: The text of the header.
    header_text = "Scientific Charts w/ Altair"


class SciChart(AltairChart):
    """
    Scientific chart with Altair.
    """

    #: The title of the chart.
    title = "Scientific Proofiness"

    def load(self) -> None:
        """
        Load the chart data.

        """
        data = pd.DataFrame({"a": list("CCCDDDEEE"), "b": [2, 7, 4, 1, 2, 6, 8, 4, 10]})
        spec = alt.Chart(data).mark_point().encode(x="a", y="b")
        self.set_data(spec)


class WidgetCodeTab(TabbedWidget):
    """
    Tabbed widget for the code behind the widget displayed in the tab.
    """

    def __init__(self, *args, **kwargs) -> None:
        title = kwargs.pop("title", "")
        code = kwargs.pop("code", [])
        widget = kwargs.pop("widget", None)
        super().__init__(*args, **kwargs)
        if not isinstance(code, list):
            code = [code]
        code_stream = WidgetStream()
        for code_block in code:
            code_widget = CodeWidget(code=code_block, language="python")
            code_stream.add_widget(code_widget)
        self.add_tab(f"{title} Widget", widget)
        self.add_tab(f"{title} Code", code_stream)


class SciSyncChartCard(CardWidget):
    """
    Scientific chart with Altair.
    """

    #: The title of the card.
    title: str = "Scientific Chart"
    #: The subtitle of the card.
    subtitle: str = "Altair"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(SciSyncChartCard.get_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_widget())
        self.set_widget(tab, "mb-5")

    def get_widget(self) -> AltairChart:
        """
        Return the scientific chart.

        Returns:
            The scientific chart.

        """
        data = pd.DataFrame({"a": list("CCCDDDEEE"), "b": [2, 7, 4, 1, 2, 6, 8, 4, 7]})
        spec = alt.Chart(data).mark_point().encode(x="a", y="b")
        chart = AltairChart(title=self.title)
        chart.set_data(spec)
        return chart


class SciChartCard(CardWidget):
    """
    Scientific chart with Altair (AJAX).
    """

    #: The title of the card.
    title: str = "Scientific Chart"
    #: The subtitle of the card.
    subtitle = "Altair (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        chart = SciChart()
        code = inspect.getsource(SciChart)
        tab = WidgetCodeTab(code=code, widget=chart)
        self.set_widget(tab, "mb-5")


class TestTable(DataTable):
    """
    Test table with DatatablesJS.
    """

    #: The model to use for the table.
    model = Measurement
    #: The ID of the table.
    table_id: str = "data_measurement"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Add all the columns.
        self.add_column("name")
        self.add_column("time", searchable=False)
        self.add_column("pressure", align="right")
        self.add_column("temperature", align="right")
        self.add_column("restricted", visible=False)
        self.add_column("open", sortable=False)

        # Set the `restricted` filter.
        filtr = DataTableFilter()
        filtr.add_choice("True", "True")
        filtr.add_choice("False", "False")
        self.add_filter("restricted", filtr)

        # Set the `open` filter.
        filtr = DataTableFilter()
        filtr.add_choice("True", "True")
        filtr.add_choice("False", "False")
        self.add_filter("open", filtr)

        # Set the `pressure` filter.
        filtr = DataTableFilter()
        filtr.add_choice("< 1000", "level_1000")
        filtr.add_choice("1000-2000", "level_2000")
        filtr.add_choice("2000-3000", "level_3000")
        self.add_filter("pressure", filtr)

    def filter_pressure_column(self, qs, column, value) -> QuerySet:  # noqa: ARG002
        """
        Filter endpoint for the ``pressure`` column.

        Args:
            qs: The queryset to filter.
            column: The column to filter (unused).
            value (str): The value to filter by.

        Returns:
            The filtered queryset.

        """
        if value == "level_1000":
            qs = qs.filter(pressure__lt=1000)
        elif value == "level_2000":
            qs = qs.filter(pressure__lt=2000).filter(pressure__gte=1000)
        elif value == "level_3000":
            qs = qs.filter(pressure__lt=3000).filter(pressure__gte=2000)
        else:
            qs = qs.filter(pressure__contains=value)
        return qs

    def filter_restricted_column(self, qs, column, value) -> QuerySet:  # noqa: ARG002
        """
        Filter endpoint for the ``restricted`` column.

        Args:
            qs: The queryset to filter.
            column: The column to filter (unused).
            value: The value to filter by.

        Returns:
            The filtered queryset.

        """
        test = value == "True"
        return qs.filter(restricted=test)

    def filter_open_column(self, qs, column, value) -> QuerySet:  # noqa: ARG002
        """
        Filter endpoint for the ``open`` column.

        Args:
            qs: The queryset to filter.
            column: The column to filter (unused).
            value (str): The value to filter by.

        Returns:
            The filtered queryset.

        """
        test = value == "True"
        return qs.filter(open=test)


class PieCard(CardWidget):
    """
    Pie chart with ChartJS.
    """

    #: The title of the card.
    title: str = "Pie Chart"
    #: The subtitle of the card.
    subtitle: str = "ChartJS"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(PieCard.get_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_widget())
        self.set_widget(tab, "mb-5")

    def get_widget(self) -> PieChart:
        """
        Return the pie chart.

        Returns:
            The pie chart.

        """
        pie = PieChart(title="Dramatic Colors")
        pie.set_categories(
            ["January", "February", "March", "April", "May", "June", "July"]
        )
        pie.add_dataset([75, 44, 92, 11, 44, 95, 35])
        return pie


class DonutCard(CardWidget):
    """
    Donut chart with ChartJS (AJAX).
    """

    #: The title of the card.
    title: str = "Donut Chart"
    #: The subtitle of the card.
    subtitle: str = "ChartJS (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        donut = TestDoughnutChart()
        code = inspect.getsource(TestDoughnutChart)
        tab = WidgetCodeTab(code=code, widget=donut)
        self.set_widget(tab, "mb-5")


class BarChartCard(CardWidget):
    """
    Bar chart with ChartJS (AJAX).
    """

    #: The title of the card.
    title: str = "Bar Chart"
    #: The subtitle of the card.
    subtitle: str = "ChartJS (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        chart = TestChart(width="500", height="400", thousands=True)
        code = inspect.getsource(TestChart)
        tab = WidgetCodeTab(code=code, widget=chart)
        self.set_widget(tab, "mb-5")


class HorizontalBarChartCard(CardWidget):
    """
    Horizontal bar chart with ChartJS.
    """

    #: The title of the card.
    title: str = "Horizontal Bar Chart"
    #: The subtitle of the card.
    subtitle: str = "ChartJS"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        code = get_code_block(HorizontalBarChartCard.get_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_widget())
        self.set_widget(tab, "mb-5")

    def get_widget(self) -> HorizontalStackedBarChart:
        """
        Return the horizontal bar chart.

        Returns:
            The horizontal bar chart.

        """
        barchart = HorizontalStackedBarChart(
            title="Formage Through July",
            money=True,
            legend=True,
            width="500",
            color=False,
        )
        barchart.set_categories(
            ["January", "February", "March", "April", "May", "June", "July"]
        )
        barchart.add_dataset([75, 44, 92, 11, 44, 95, 35], "Central")
        barchart.add_dataset([41, 92, 18, 35, 73, 87, 92], "Eastside")
        barchart.add_dataset([87, 21, 94, 13, 90, 13, 65], "Westside")
        return barchart


class HistogramCard(CardWidget):
    """
    Histogram with ChartJS (AJAX).
    """

    #: The title of the card.
    title: str = "Histogram"
    #: The subtitle of the card.
    subtitle: str = "ChartJS (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        histogram = TestHistogram(width="500", height="400")
        code = inspect.getsource(TestHistogram)
        tab = WidgetCodeTab(code=code, widget=histogram)
        self.set_widget(tab, "mb-5")


class HorizontalHistogramCard(CardWidget):
    """
    Horizontal histogram with ChartJS.
    """

    #: The title of the card.
    title: str = "Horizontal Histogram"
    #: The subtitle of the card.
    subtitle: str = "ChartJS"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        histogram = TestHorizontalHistogram(width="500", height="400")
        code = inspect.getsource(TestHorizontalHistogram)
        tab = WidgetCodeTab(code=code, widget=histogram)
        self.set_widget(tab, "mb-5")


class TableHeader(BasicHeader):
    """
    Header for the scientific tables with DatatablesJS page.
    """

    #: The level of the header.
    header_level = 5
    #: The type of the header.
    header_type = "display"
    #: The text of the header.
    header_text = "Scientific Tables w/ DatatablesJS"
    #: The CSS class of the header.
    css_class = ""


class StaticTable(StaticTableWidget):
    """
    Static data table build with :class:`~wildewidgets.StaticTableWidget`.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add_heading("Time")
        self.add_heading("Pressure")
        self.add_heading("Temperature")
        self.add_row([12, 53, 25])
        self.add_row([13, 63, 24])
        self.add_row([14, 73, 23])


class StaticTableCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.StaticTable` widget.
    """

    #: The introduction text.
    INTRO: str = """
This table is built with :class:`~wildewidgets.StaticTableWidget`.
It is a static table that does not use dataTables.js useful for small lists of data,
especially data that is not in Django models.
"""

    #: The title of the card.
    title: str = "Static Table"
    #: The subtitle of the card.
    subtitle: str = "Static Table"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = inspect.getsource(StaticTable)
        text = Block(self.INTRO, css_class="mb-4")
        tab = WidgetCodeTab(code=code, widget=StaticTable())
        self.set_widget(Block(text, tab), "mb-5")


class DataTableCard(CardWidget):
    """
    Static Data table with DatatablesJS (Non-AJAX).
    """

    #: The title of the card.
    title: str = "Data Table"
    #: The subtitle of the card.
    subtitle: str = "DatatablesJS (Non-AJAX)"

    #: The introduction text.
    INTRO: str = """
This table is built on a lower level class that provides a lot of flexibility,
but can often require more code to specify.  You can use any data source as
appropriate. Here, we are manually adding data in rows.
"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(DataTableCard.get_widget)
        text = Block(self.INTRO, css_class="mb-4")
        tab = WidgetCodeTab(code=code, widget=self.get_widget())
        self.set_widget(Block(text, tab), "mb-5")

    def get_widget(self) -> DataTable:
        """
        Return the data table.

        Returns:
            The data table.

        """
        table = DataTable()
        table.is_data_list = False
        table.add_column("time")
        table.add_column("pressure")
        table.add_column("temperature")
        table.add_row(time=12, pressure=53, temperature=25)
        table.add_row(time=13, pressure=63, temperature=24)
        table.add_row(time=14, pressure=73, temperature=23)
        return table


class TestTableCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.TestTable` widget.
    """

    #: The title of the card.
    title: str = "Test Table"
    #: The subtitle of the card.
    subtitle: str = "DatatablesJS (AJAX)"
    #: The introduction text.
    INTRO: str = """
This table is built on a lower level class that provides a lot of flexibility,
but can often require more code to specify. You can use any data source as
appropriate. Here, we are using a Django Model.
"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = inspect.getsource(TestTable)
        text = Block(self.INTRO, css_class="mb-4")
        tab = WidgetCodeTab(code=code, widget=TestTable())
        self.set_widget(Block(text, tab), "mb-5")


class BookModelTable(BasicModelTable):
    """
    Book model table with DatatablesJS (AJAX).
    """

    #: The model to use for the table.
    model: type[Model] = Book
    #: The fields to display in the table.
    fields: list[str] = ["title", "authors__full_name", "isbn"]  # noqa: RUF012
    #: The alignment of the columns.
    alignment: ClassVar[dict[str, Literal["left", "right", "center"]]] = {  # type: ignore[misc]
        "authors": "left",
    }
    #: The verbose names of the columns.
    verbose_names: dict[str, str] = {"authors__full_name": "Authors"}  # noqa: RUF012
    #: Whether to show the dataTables.js buttons
    buttons: bool = True
    #: Whether to stripe the table rows.
    striped: bool = True

    def render_authors__full_name_column(self, row, column):  # noqa: ARG002
        """
        Return the full name of the authors, split into multiple lines if there
        are multiple authors.

        Args:
            row: The row to render.
            column: The column to render.

        Returns:
            The full name of the authors.

        """
        authors = row.authors.all()
        if authors.count() > 1:
            return f"{authors[0].full_name} ... "
        return authors[0].full_name


class BookModelTableCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.BookModelTable` widget.
    """

    #: The introduction text.
    INTRO: str = """
This table is built on a high level class designed to work easily with Django
models. It automatically generates the table columns and provides a lot of
flexibility for customization.  It is dynamic, using AJAX to retrieve data from
the server.
"""
    title = "Book Model Table"
    subtitle = "DatatablesJS (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = inspect.getsource(BookModelTable)
        text = Block(self.INTRO, css_class="mb-4")
        tab = WidgetCodeTab(code=code, widget=BookModelTable())
        self.set_widget(Block(text, tab), "mb-5")


class PressureCellWidget(Block):
    """
    Pressure cell widget with FontIcon.  This is used to demonstrate the use of
    cell widgets in a table.
    """

    def __init__(self, *args, row=None, column="", **kwargs):  # noqa: ARG002
        value = row.pressure
        if value > 2000:  # noqa: PLR2004
            icon = "thermometer-high"
            color = "red"
        elif value > 1000:  # noqa: PLR2004
            icon = "thermometer-half"
            color = "orange"
        else:
            icon = "thermometer-low"
            color = "green"
        super().__init__(
            HorizontalLayoutBlock(
                FontIcon(icon=icon, color=color), f"{value}", justify="end"
            ),
            *args,
            **kwargs,
        )


class WidgetCellTable(WidgetCellMixin, BasicModelTable):  # type: ignore[misc]
    """
    Table with widget cells.  This is used to demonstrate the use of widget cells
    in a table.
    """

    #: The fields to display in the table.
    fields: list[str] = ["name", "pressure", "temperature"]  # noqa: RUF012
    #: The cell widgets to use for the table.
    cell_widgets: dict[str, type[Widget]] = {"pressure": PressureCellWidget}  # noqa: RUF012
    #: The model to use for the table.
    model: type[Model] = Measurement
    #: The alignment of the columns.
    alignment: ClassVar[dict[str, Literal["left", "right", "center"]]] = {  # type: ignore[misc]
        "pressure": "right",
        "temperature": "right",
    }
    #: Whether to stripe the table rows.
    striped: bool = True


class WidgetCellTableCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.WidgetCellTable` widget.
    """

    #: The title of the card.
    title: str = "Widget Cell Table"
    #: The subtitle of the card.
    subtitle: str = "DatatablesJS (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = []
        code.append(inspect.getsource(PressureCellWidget))
        code.append(inspect.getsource(WidgetCellTable))
        tab = WidgetCodeTab(code=code, widget=WidgetCellTable())
        self.set_widget(tab, "mb-5")


class ApexHeader(BasicHeader):
    """
    Header for the Apex Charts page.
    """

    #: The level of the header.
    header_level = 5
    #: The type of the header.
    header_type = "display"
    #: The text of the header.
    header_text = "Apex Charts"
    #: The CSS class of the header.
    css_class = ""


class ApexLineDataset(ApexDatasetBase):
    """
    Apex line dataset.
    """

    #: The name of the dataset.
    name: str = "Data"

    def __init__(self, *args, **kwargs):  # noqa: ARG002
        super().__init__(**kwargs)
        self.data = [25, 66, 41, 89, 63, 25, 44, 12, 36, 9, 54]


class ApexLineChart(ApexJSONMixin, ApexChartBase):
    """
    Apex line chart.
    """

    #: The type of the chart.
    chart_type: str = "line"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_categories(
            [
                "Jan",
                "Feb",
                "Mar",
                "Apr",
                "May",
                "Jun",
                "Jul",
                "Aug",
                "Sep",
                "Oct",
                "Nov",
                "Dec",
            ]
        )

    def load(self) -> None:
        """
        Load the Apex line dataset.
        """
        self.add_dataset(ApexLineDataset())


class ApexChart1Card(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.ApexLineChart` widget.
    """

    #: The title of the card.
    title: str = "Apex Charts"
    #: The subtitle of the card.
    subtitle: str = "Line Chart (AJAX)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = []
        code.append(inspect.getsource(ApexLineDataset))
        code.append(inspect.getsource(ApexLineChart))
        tab = WidgetCodeTab(code=code, widget=ApexLineChart())
        self.set_widget(tab, "mb-5")


class ApexSparkLineChart(ApexSparkline):
    """
    Apex spark line chart.
    """

    #: The type of the chart.
    chart_type: str = "bar"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_dataset(ApexLineDataset())
        self.add_categories(
            [
                "Jan",
                "Feb",
                "Mar",
                "Apr",
                "May",
                "Jun",
                "Jul",
                "Aug",
                "Sep",
                "Oct",
                "Nov",
                "Dec",
            ]
        )


class ApexSparklineCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.ApexSparkLineChart` widget.
    """

    #: The title of the card.
    title: str = "Apex Charts"
    #: The subtitle of the card.
    subtitle: str = "Sparkline"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = []
        code.append(inspect.getsource(ApexLineDataset))
        code.append(inspect.getsource(ApexSparkLineChart))
        tab = WidgetCodeTab(code=code, widget=ApexSparkLineChart())
        self.set_widget(tab, "mb-5")


class MarkdownCard(CardWidget):
    """
    Demonstration of the :class:`~wildewidgets.MarkdownWidget` widget.
    """

    #: The markdown text.
    MARKDOWN: str = """
# Markdown Example

Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu
fugiat nulla pariatur.

    Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia
    deserunt mollit anim id est laborum.

## de Finibus Bonorum et Malorum

1. Sed ut perspiciatis unde omnis iste natus error sit voluptatem accusantium
   doloremque laudantium, totam rem aperiam, eaque ipsa quae ab illo inventore
   veritatis et quasi architecto beatae vitae dicta sunt explicabo.

2. Nemo enim ipsam voluptatem quia voluptas sit aspernatur aut odit aut fugit,
   sed quia...
"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(MarkdownCard.get_markdown_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_markdown_widget())
        self.set_widget(tab, "mb-5")

    def get_markdown_widget(self) -> MarkdownWidget:
        """
        Return the Markdown widget.
        """
        return MarkdownWidget(text=self.MARKDOWN)


class HTMLCard(CardWidget):
    """
    Demonstration of the :class:`~wildewidgets.HTMLWidget` widget.
    """

    #: The HTML text.
    HTML: str = """
<h1>HTML Example</h1>
<p>Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod
tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam,
quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo
consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum
dolore eu fugiat nulla pariatur. </p>

<pre><code>Excepteur sint occaecat cupidatat non proident, sunt in culpa qui
officia deserunt mollit anim id est laborum.  </code></pre>

<h2>de Finibus Bonorum et Malorum</h2>
<ol>
<li>
<p>Sed ut perspiciatis unde omnis iste natus error sit voluptatem accusantium
doloremque laudantium, totam rem aperiam, eaque ipsa quae ab illo inventore
veritatis et quasi architecto beatae vitae dicta sunt explicabo.</p>
</li>
<li>
<p>Nemo enim ipsam voluptatem quia voluptas sit aspernatur aut odit aut fugit,
sed quia...</p>
</li>
</ol>
"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(HTMLCard.get_html_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_html_widget())
        self.set_widget(tab, "mb-5")

    def get_html_widget(self) -> HTMLWidget:
        """
        Return the HTML widget.
        """
        return HTMLWidget(html=self.HTML)


class CodeCard(CardWidget):
    """
    Demonstration of the :class:`~wildewidgets.CodeWidget` widget.
    """

    #: The title of the card.
    title: str = "Code"
    #: The subtitle of the card.
    subtitle: str = "Code Widget"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(CodeCard.get_code_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_code_widget())
        self.set_widget(tab, "mb-5")

    def get_code_widget(self) -> CodeWidget:
        """
        Return the Code widget.
        """
        return CodeWidget(code=inspect.getsource(MarkdownCard), language="python")


class StringCard(CardWidget):
    """
    Demonstration of these widgets:

    - :class:`~wildewidgets.StringBlock`
    - :class:`~wildewidgets.TimeStampBlock`
    - :class:`~wildewidgets.LabelBlock`
    - :class:`~wildewidgets.TagBlock`
    """

    #: The title of the card.
    title: str = "Various String Blocks"
    #: The icon of the card.
    icon: str = "fonts"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(StringCard.get_string_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_string_widget())
        self.set_widget(tab, "mb-5")

    def get_string_widget(self) -> Block:
        """
        Return the String widget.
        """
        return Block(
            Block("This is a string block", css_class="mb-2"),
            TimeStamp("4:53 PM - this is a time block."),
            Block(LabelBlock("This is a label block", css_class="my-2")),
            TagBlock("This is a tag block"),
        )


class BookModelWidget(Block):
    """
    Simple demonstration of using a :class:`~wildewidgets.Block` to
    display rows in a :class:`~wildewidgets.PagedModelWidget`.
    """

    #: The object to display.
    object: Model | None = None

    def __init__(self, object: Model | None = None, **kwargs) -> None:  # noqa: A002
        if object is None:
            msg = "object must be provided"
            raise ValueError(msg)
        authors = ", ".join([x.full_name for x in object.authors.all()])
        super().__init__(
            Block(
                f"{object.title} - {authors}", css_class="p-3 mb-1 border text-bg-light"
            ),
            **kwargs,
        )


class PagedBookWidget(PagedModelWidget):
    """
    Paged demonstration of using a :class:`~wildewidgets.PagedModelWidget` to
    display a list of objects.
    """

    #: The model to use for the widget.
    model = Book
    #: The widget to use for each object.
    model_widget = BookModelWidget
    #: The number of objects to display per page.
    paginate_by: int = 5


class PagedBookCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.PagedBookWidget` widget.
    """

    #: The title of the card.
    title: str = "Paged Model Widget"
    #: The icon of the card.
    icon: str = "book"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = []
        code.append(inspect.getsource(BookModelWidget))
        code.append(inspect.getsource(PagedBookWidget))
        tab = WidgetCodeTab(code=code, widget=self.get_paged_book_widget())
        self.set_widget(tab, "mb-5")

    def get_paged_book_widget(self) -> PagedBookWidget:
        return PagedBookWidget()


class AuthorListCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.AuthorListModelWidget` widget.
    """

    #: The title of the card.
    title: str = "List Model Widget"
    #: The icon of the card.
    icon: str = "people"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(AuthorListCard.get_author_list_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_author_list_widget())
        self.set_widget(tab, "mb-5")

    def get_author_list_widget(self) -> ListModelWidget:
        """
        Return the ListModelWidget.
        """
        return ListModelWidget(queryset=Author.objects.all()[:10])


class AuthorListModelCardWidgetCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.ListModelCardWidget` widget.
    """

    #: The title of the card.
    title: str = "List Model Card Widget"
    #: The icon of the card.
    icon = "person-square"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(AuthorListModelCardWidgetCard.get_author_list_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_author_list_widget())
        self.set_widget(tab, "mb-5")

    def get_author_list_widget(self) -> ListModelCardWidget:
        """
        Return the ListModelCardWidget.
        """
        return ListModelCardWidget(queryset=Author.objects.all()[:10])


class TabCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.TabbedWidget` widget.
    """

    #: The title of the card.
    title: str = "Tab Widget"
    #: The icon of the card.
    icon: str = "tab"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(TabCard.get_tab_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_tab_widget())
        self.set_widget(tab, "mb-5")

    def get_tab_widget(self) -> TabbedWidget:
        """
        Return the TabbedWidget with two tabs.
        """
        tab = TabbedWidget()
        tab.add_tab("Tab 1", Block("This is tab 1"))
        tab.add_tab("Tab 2", Block("This is tab 2"))
        return tab


class CardCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.CardWidget` widget.
    """

    #: The title of the card.
    title: str = "Card Widget"
    #: The icon of the card.
    icon: str = "card"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(CardCard.get_card_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_card_widget())
        self.set_widget(tab, "mb-5")

    def get_card_widget(self) -> CardWidget:
        """
        Return the CardWidget.
        """
        card = CardWidget(header_text="This is the card's header")
        card.set_widget(Block("This is a the card's content."))
        return card


class CrispyFormCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.CrispyFormWidget` widget.
    """

    #: The title of the card.
    title: str = "Crispy Form Widget"
    #: The icon of the card.
    icon: str = "form"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(CrispyFormCard.get_crispy_form_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_crispy_form_widget())
        self.set_widget(tab, "mb-5")

    def get_crispy_form_widget(self) -> CrispyFormWidget:
        """
        Return the CrispyFormWidget.
        """
        form = MinimalBookForm()
        form.helper.form_method = "get"
        return CrispyFormWidget(form=form)


class CollapseCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.CollapseWidget` widget.
    """

    #: The title of the card.
    title: str = "Collapse Widget"
    #: The icon of the card.
    icon: str = "collapse"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(CollapseCard.get_collapse_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_collapse_widget())
        self.set_widget(tab, "mb-5")

    def get_collapse_widget(self) -> Block:
        """
        Return the CollapseWidget.
        """
        header = CardHeader(header_text="Collapse Widget")
        header.add_collapse_button(
            text="Show Form",
            color="primary",
            target="#collapse1",
        )
        form = MinimalBookForm()
        form.helper.form_method = "get"
        form_widget = CollapseWidget(
            CrispyFormWidget(form=form),
            css_id="collapse1",
        )
        return Block(header, form_widget)


class HorizontalLayoutCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.HorizontalLayoutBlock` widget.
    """

    #: The title of the card.
    title: str = "Horizontal Layout Widget"
    #: The icon of the card.
    icon: str = "layout"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(HorizontalLayoutCard.get_horizontal_layout_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_horizontal_layout_widget())
        self.set_widget(tab, "mb-5")

    def get_horizontal_layout_widget(self) -> HorizontalLayoutBlock:
        return HorizontalLayoutBlock(
            Block("This is a string block", css_class="border p-3 bg-azure"),
            Block("This is a string block", css_class="border p-3 bg-yellow"),
            Block("This is a string block", css_class="border p-3 bg-lime"),
            Block("This is a string block", css_class="border p-3 bg-cyan"),
        )


class ModalCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.ModalWidget` widget.
    """

    #: The title of the card.
    title: str = "Modal Widget"
    #: The icon of the card.
    icon: str = "modal"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(ModalCard.get_modal_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_modal_widget())
        self.set_widget(tab, "mb-5")

    def get_modal_widget(self) -> Block:
        """
        Return the ModalWidget.
        """
        header = CardHeader(header_text="Modal Widget")
        header.add_modal_button(
            text="Show Modal",
            color="primary",
            target="#modal1",
        )
        modal = ModalWidget(
            modal_id="modal1",
            modal_title="Modal Title",
            modal_body="This is a modal widget",
        )
        return Block(header, modal)


class CrispyFormModalCard(CardWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.CrispyFormModalWidget` widget.
    """

    #: The title of the card.
    title: str = "Crispy Form Modal Widget"
    #: The icon of the card.
    icon: str = "form"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        code = get_code_block(CrispyFormModalCard.get_crispy_form_modal_widget)
        tab = WidgetCodeTab(code=code, widget=self.get_crispy_form_modal_widget())
        self.set_widget(tab, "mb-5")

    def get_crispy_form_modal_widget(self) -> Block:
        """
        Return the CrispyFormModalWidget.
        """
        header = CardHeader(header_text="Crispy Form Modal Widget")
        header.add_modal_button(
            text="Show Modal",
            color="primary",
            target="#modal2",
        )
        form = MinimalBookForm()
        form.helper.form_method = "get"
        modal = CrispyFormModalWidget(
            modal_id="modal2",
            modal_title="Modal Title",
            form=form,
        )
        return Block(header, modal)


class HomeTable(BasicModelTable):
    """
    Demonstration of the :class:`~demo.core.wildewidgets.BasicModelTable` widget.
    """

    #: The model of the table.
    model: type[Model] = Book
    #: The fields of the table.
    fields: ClassVar[list[str]] = ["title", "isbn", "binding"]  # type: ignore[misc]
    #: Whether to hide the controls.
    hide_controls: bool = True
    #: The page length of the table.
    page_length: int = 25
    #: Whether to stripe the table.
    striped: bool = True
    #: Whether to make the table small.
    small: bool = True
    #: Whether to make the table unsearchable.
    unsearchable: ClassVar[list[str]] = ["binding"]  # type: ignore[misc]


class AuthorListModelWidget(ListModelWidget):
    """
    Wrapper around the :class:`~demo.core.wildewidgets.ListModelWidget` widget.
    """

    def get_object_text(self, obj) -> str:
        """
        Return the object text.
        """
        return f"{FontIcon(icon='person-fill', css_class='pe-2')!s} {obj!s}"


class HomeBlock(Block):
    """
    Main block for the demo home page.
    """

    #: The description of the demo project
    description = """
django-wildewidgets is a Django design library providing several tools for
building full-featured, widget-based web applications with a standard,
consistent design, based on Bootstrap.
"""
    #: The features of wildewidgets.
    features: ClassVar[list[str]] = [
        "Large library of standard widgets",
        "Custom widgets",
        "Widgets can be composable",
        "Teplateless design",
        "AJAX data for tables, charts, and other data based widgets",
        "Several supporting views",
    ]
    #: The standard widgets of wildewidgets.
    standard_widgets: ClassVar[list[str]] = [
        "Basic blocks",
        "Template based widgets",
        "Basic Buttons",
        "Form, Modal, and Collapse Buttons",
        "Header widgets",
        "Chart widgets, including Altair, Apex, and ChartJS",
        "Layout and structural widgets, like Card and Tab widgets",
        "Modal widgets",
        "Form widgets",
        "Table widgets",
        "Text widgets, like HTML, Code, Markdown, and Label widgets",
        "Other miscillaneous widgets, like Breadcrumb, "
        "Gravatar, and KeyValueList widgets.",
    ]
    #: Footer notice for the demo project.
    notice: str = (
        "This entire site is built using django-wildewidgets, "
        "with only a single shared template."
    )

    def __init__(self, *args, **kwargs):  # noqa: ARG002
        header = PageHeader(
            header_text="django-wildewidgets Demo",
            css_class=f"{PageHeader.css_class} mx-1",
        )

        pie = PieChart(width=100, height=100)
        pie.set_categories(
            ["January", "February", "March", "April", "May", "June", "July"]
        )
        pie.add_dataset([75, 44, 92, 11, 44, 95, 35])

        modal = ModalWidget(
            modal_id="modal1",
            modal_title="Modal Title",
            modal_body=Block("This is a modal widget"),
        )

        super().__init__(
            header,
            CardWidget(
                widget=Block(
                    Block(self.description, css_class="w-75"),
                    Block(self.notice, css_class="w-75 my-3 fw-bold"),
                ),
                css_class="mb-2",
            ),
            HorizontalLayoutBlock(
                Block(
                    CardWidget(
                        header=BasicHeader(
                            header_text="Features", css_class="my-0", header_level=3
                        ),
                        widget=UnorderedList(*self.features),
                        css_class="mb-2",
                    ),
                    CardWidget(
                        header=BasicHeader(
                            header_text="Standard Widgets",
                            css_class="my-0",
                            header_level=3,
                        ),
                        widget=UnorderedList(*self.standard_widgets),
                        css_class="mb-2",
                    ),
                ),
                Block(
                    CardWidget(
                        widget=TestDoughnutChart(legend=False, width=240, height=300),
                        css_class="mb-2",
                    ),
                    CardWidget(
                        widget=HorizontalLayoutBlock(
                            ModalButton(
                                text=Block(
                                    FontIcon(icon="window", css_class="pe-1"), "Modal"
                                ),
                                color="primary",
                                target="#modal1",
                            ),
                            LinkButton(
                                url="https://github.com/caltechads/django-wildewidgets",
                                text=Block(
                                    TablerFontIcon(
                                        icon="brand-github", css_class="pe-1"
                                    ),
                                    "Github",
                                ),
                                color="secondary",
                            ),
                            justify="evenly",
                        ),
                        css_class="py-3 mb-2",
                    ),
                    CardWidget(
                        widget=AuthorListModelWidget(
                            queryset=Author.objects.all()[5:8]
                        ),
                        css_class="mb-2",
                        attributes={"style": "height: 223px;"},
                    ),
                    css_class="mx-xl-2",
                ),
                Block(
                    CardWidget(
                        widget=TestChart(width="300px", height="371px"),
                        css_class="mb-2",
                    ),
                    CardWidget(
                        widget=ApexLineChart(),
                        css_class="mb-2 pt-4",
                    ),
                ),
                align="start",
                flex_size="xl",
            ),
            CardWidget(
                widget=HomeTable(),
                css_class=" overflow-auto",
            ),
            modal,
        )



import uuid

from ..base import Block

# -----------------------------------------------------------------------------
# Widgets
# -----------------------------------------------------------------------------


class StaticTableRowWidget(Block):
    """
    Simple static table row widget.  This widget is used to create a row in a
    :class:`~workshop_admin.core.wildewidgets.StaticTableWidget`.

    Args:
        cells: the cells to add to the row
        cell_tag: the HTML tag for the cells
        cell_css_class: the CSS class for the cells
        **kwargs: additional keyword arguments to pass to the block

    """

    #: The HTML tag for the row
    tag: str = "tr"
    #: The HTML tag for the cells
    cell_tag: str = "td"

    def __init__(
        self,
        cells: list[str | Block],
        cell_tag: str | None = None,
        cell_css_class: str | None = None,
        **kwargs,
    ) -> None:
        if cell_tag is not None:
            self.cell_tag = cell_tag
        rendered_cells: list[Block] = []
        for cell in cells:
            if isinstance(cell, Block) and cell.tag == self.cell_tag:
                rendered_cells.append(cell)
                continue
            rendered_cells.append(
                Block(cell, tag=self.cell_tag, css_class=cell_css_class)
            )
        super().__init__(*rendered_cells, **kwargs)


class StaticTableWidget(Block):
    """
    Simple static table widget.  This differs from the
    :class:`~wildewidgets.DataTable` type widgets in that it does not use
    dataTables.js and is therefore simpler for small lists of data.

    It offers a sorting mechanism by clicking on the "Sort" button in the heading
    of the column.  The table is sorted by the text content of the cells in the
    column that the button is in.  If every cell in that column is numeric
    (optional minus, digits, optional decimal), the column is sorted numerically;
    otherwise locale-aware alphabetical sort is used.  Empty cells in numeric
    columns sort last (ascending) or first (descending).

    Examples:
        With constructor arguments:

        .. code-block:: python

            from wildewidgets import StaticTableWidget

            table = StaticTableWidget(
                headings=["Name", "Age", "City"],
                rows=[["John", 25, "New York"], ["Jane", 30, "Los Angeles"]]
            )

            print(table)

        With constructor overrides:

        .. code-block:: python

            from wildewidgets import StaticTableWidget

            class MyStaticTableWidget(StaticTableWidget):
                def __init__(self, *args, **kwargs):
                    super().__init__(*args, **kwargs)
                    self.add_heading("Name")
                    self.add_heading("Age")
                    self.add_heading("City")
                    self.add_row(["John", 25, "New York"])
                    self.add_row(["Jane", 30, "Los Angeles"])

            table = MyStaticTableWidget()
            print(table)

        With method calls:

        .. code-block:: python

            from wildewidgets import StaticTableWidget

            table = StaticTableWidget()
            table.add_heading("Name")
            table.add_heading("Age")
            table.add_heading("City")
            table.add_row(["John", 25, "New York"])
            table.add_row(["Jane", 30, "Los Angeles"])

            print(table)


        .. code-block:: python

            from wildewidgets import StaticTableWidget
            table = StaticTableWidget(
                headings=["Name", "Age", "City"],
                rows=[["John", 25, "New York"], ["Jane", 30, "Los Angeles"]]
            )
            table.add_heading("Country")
            table.add_heading("Occupation")
            table.add_heading("City"sjj
            table.add_row(
                 ["John", 25, "New York", "United States", "Software Engineer"]
            )
            table.add_row(
                ["Jane", 30, "Los Angeles", "United States", "Software Engineer"]
            )

            print(table)


    Keyword Args:
        headings: the headings for the table
        rows: the rows for the table
        cell_css_class: the CSS class for the cells
        **kwargs: additional keyword arguments to pass to the block

    """

    #: The HTML tag for the table
    tag: str = "table"
    #: The HTML name for the widget for CSS styling and JavaScript
    name: str = "static-table"

    @property
    def script(self) -> str:
        """
        A JavaScript script to sort the table by clicking on a button in the
        headings.

        Iniitially, the table is sorted in the order they are listed in the data
        source.

        It draws a button in the ``<th>`` elements named "Sort" with the current
        sort order represented by an arrow (if we are sorting) or a blank (if we
        are not sorting).  When the button is clicked, the table is sorted by
        the opposite of the current sort order or ascending (if no sort order is
        set), and the button text is updated to the opposite of the selected
        sort order.

        Note:
            The table is sorted by the text content of the cells in the column
            that the button is in.  If every cell in that column is numeric
            (optional minus, digits, optional decimal), the column is sorted
            numerically; otherwise locale-aware alphabetical sort is used.
            Empty cells in numeric columns sort last (ascending) or first
            (descending).

        Args:
            self: the widget instance

        Returns:
            A string of JavaScript code to sort the table by clicking on the headings.

        """
        sort_button = '<button class="btn btn-link btn-sm" style="font-size: 0.8em;">Sort</button>'  # noqa: E501
        return f"""
        document.addEventListener("DOMContentLoaded", function() {{
            var table = document.querySelector('table#{self._css_id}');
            // Add the up and down arrows to the headings.
            var headings = table.querySelectorAll('thead th');
            headings.forEach(function(heading) {{
                heading.innerHTML = heading.innerHTML + '{sort_button}';
            }});
            function isNumericColumn(rows, colIdx) {{
                var numericPattern = /^-?\\d*\\.?\\d+$/;
                for (var i = 0; i < rows.length; i++) {{
                    var cell = rows[i].cells[colIdx];
                    if (!cell) return false;
                    var text = (cell.textContent || '').trim();
                    if (text !== '' && !numericPattern.test(text)) return false;
                }}
                return true;
            }}
            function compareCells(aVal, bVal, numeric, asc) {{
                if (numeric) {{
                    var na = parseFloat(aVal);
                    var nb = parseFloat(bVal);
                    var aEmpty = (aVal.trim() === '') || isNaN(na);
                    var bEmpty = (bVal.trim() === '') || isNaN(nb);
                    if (aEmpty && bEmpty) return 0;
                    if (aEmpty) return asc ? 1 : -1;
                    if (bEmpty) return asc ? -1 : 1;
                    return asc ? (na - nb) : (nb - na);
                }}
                return asc ? aVal.localeCompare(bVal) : bVal.localeCompare(aVal);
            }}
            if (table) {{
                table.addEventListener('click', function(event) {{
                    var target = event.target;
                    if (target.tagName === 'BUTTON') {{
                        var columnIndex = target.parentElement.cellIndex;
                        var is_ascending = target.textContent.includes('↓');
                        var rows = Array.from(table.querySelectorAll('tbody tr'));
                        var numeric = isNumericColumn(rows, columnIndex);
                        if (!is_ascending) {{
                            rows.sort(function(a, b) {{
                                var aText = (a.cells[columnIndex].textContent || '').trim();
                                var bText = (b.cells[columnIndex].textContent || '').trim();
                                return compareCells(aText, bText, numeric, true);
                            }});
                        }} else {{
                            rows.sort(function(a, b) {{
                                var aText = (a.cells[columnIndex].textContent || '').trim();
                                var bText = (b.cells[columnIndex].textContent || '').trim();
                                return compareCells(aText, bText, numeric, false);
                            }});
                        }}
                        var tbody = table.querySelector('tbody');
                        tbody.innerHTML = '';
                        rows.forEach(function(row) {{
                            tbody.appendChild(row);
                        }});
                        // update the button text to the opposite of the current
                        // sort order
                        target.textContent = "Sort " + (!is_ascending ? '↓' : '↑');
                        // Set the buttons in the other columns to no sort
                        var otherColumns = table.querySelectorAll('thead th');
                        otherColumns.forEach(function(column) {{
                            if (column !== target.parentElement) {{
                                column.querySelector('button').textContent = 'Sort';
                            }}
                        }});
                    }}
                }});
            }}
        }});
        """  # noqa: E501, S608

    @script.setter
    def script(self, value: str) -> None:
        self._script = value

    def __init__(
        self,
        headings: list[str | Block] | None = None,
        rows: list[list[str | Block]] | None = None,
        cell_css_class: str | None = None,
        **kwargs,
    ) -> None:
        if headings is not None and rows is not None:
            assert len(headings) == len(rows[0]), (  # noqa: S101
                "Headings and rows must have the same length"
            )
        super().__init__(css_class="table table-sm mb-0", **kwargs)
        self.headings = headings or []
        self.rows = rows or []
        self.header = Block(tag="thead")
        self.body = Block(tag="tbody")
        self.cell_css_class = cell_css_class

        if self.headings:
            for heading in self.headings:
                self.header.add_block(
                    Block(heading, tag="th", css_class=self.cell_css_class)
                )
        for row in self.rows:
            self.body.add_block(
                StaticTableRowWidget(row, cell_css_class=self.cell_css_class)
            )

        if self.headings:
            self.add_block(self.header)
        if self.rows:
            self.add_block(self.body)

        if not self._css_id:
            self._css_id = f"static-table-{uuid.uuid4()}"

    def add_heading(self, heading: str | Block) -> None:
        """
        Add a single heading to the table.  Do this before adding any rows.

        Args:
            heading: the heading to add to the table

        """
        self.header.add_block(Block(heading, tag="th", css_class=self.cell_css_class))
        if self.header not in self.blocks:
            self.add_block(self.header)

    def add_row(self, row: list[str | Block]) -> None:
        """
        Add a single row to the table.  Do this after adding headings.
        """
        self.body.add_block(
            StaticTableRowWidget(row, cell_css_class=self.cell_css_class)
        )
        if self.body not in self.blocks:
            self.add_block(self.body)
